# encoding: utf-8


def upgrade(migrate_engine):
    migrate_engine.execute('''
        ALTER TABLE package_revision
            ADD COLUMN metadata_created timestamp without time zone;
        ALTER TABLE package
            ADD COLUMN metadata_created timestamp without time zone;

        UPDATE package SET metadata_created=
            (SELECT revision_timestamp
             FROM package_revision
             WHERE id=package.id
             ORDER BY revision_timestamp ASC
             LIMIT 1);
    ''')



# encoding: utf-8

from logging import getLogger

import ckan.plugins as p

log = getLogger(__name__)

class StatsPlugin(p.SingletonPlugin):
    '''Stats plugin.'''

    p.implements(p.IRoutes, inherit=True)
    p.implements(p.IConfigurer, inherit=True)

    def after_map(self, map):
        map.connect('stats', '/stats',
            controller='ckanext.stats.controller:StatsController',
            action='index')
        map.connect('stats_action', '/stats/{action}',
            controller='ckanext.stats.controller:StatsController')
        return map

    def update_config(self, config):
        templates = 'templates'
        if p.toolkit.asbool(config.get('ckan.legacy_templates', False)):
                templates = 'templates_legacy'
        p.toolkit.add_template_directory(config, templates)
        p.toolkit.add_public_directory(config, 'public')
        p.toolkit.add_resource('public/ckanext/stats', 'ckanext_stats')



# encoding: utf-8

import sys
import logging
import re

import sqlalchemy.engine.url as sa_url

import ckan.plugins as p
import ckan.logic as logic
import ckan.model as model
from ckan.common import config
import ckanext.datastore.logic.action as action
import ckanext.datastore.logic.auth as auth
import ckanext.datastore.db as db
import ckanext.datastore.interfaces as interfaces
import ckanext.datastore.helpers as datastore_helpers
from ckanext.datastore.helpers import literal_string


log = logging.getLogger(__name__)
_get_or_bust = logic.get_or_bust

DEFAULT_FORMATS = []

ValidationError = p.toolkit.ValidationError


def _is_legacy_mode(config):
    '''
        Decides if the DataStore should run on legacy mode

        Returns True if `ckan.datastore.read_url` is not set in the provided
        config object or CKAN is running on Postgres < 9.x
    '''
    write_url = config.get('ckan.datastore.write_url')

    engine = db._get_engine({'connection_url': write_url})
    connection = engine.connect()

    return (not config.get('ckan.datastore.read_url') or
            not db._pg_version_is_at_least(connection, '9.0'))


class DatastoreException(Exception):
    pass


class DatastorePlugin(p.SingletonPlugin):
    p.implements(p.IConfigurable, inherit=True)
    p.implements(p.IActions)
    p.implements(p.IAuthFunctions)
    p.implements(p.IResourceUrlChange)
    p.implements(p.IDomainObjectModification, inherit=True)
    p.implements(p.IRoutes, inherit=True)
    p.implements(p.IResourceController, inherit=True)
    p.implements(interfaces.IDatastore, inherit=True)

    legacy_mode = False
    resource_show_action = None

    def __new__(cls, *args, **kwargs):
        idatastore_extensions = p.PluginImplementations(interfaces.IDatastore)
        idatastore_extensions = idatastore_extensions.extensions()

        if idatastore_extensions and idatastore_extensions[0].__class__ != cls:
            msg = ('The "datastore" plugin must be the first IDatastore '
                   'plugin loaded. Change the order it is loaded in '
                   '"ckan.plugins" in your CKAN .ini file and try again.')
            raise DatastoreException(msg)

        return super(cls, cls).__new__(cls, *args, **kwargs)

    def configure(self, config):
        self.config = config
        # check for ckan.datastore.write_url and ckan.datastore.read_url
        if (not 'ckan.datastore.write_url' in config):
            error_msg = 'ckan.datastore.write_url not found in config'
            raise DatastoreException(error_msg)

        # Legacy mode means that we have no read url. Consequently sql search is not
        # available and permissions do not have to be changed. In legacy mode, the
        # datastore runs on PG prior to 9.0 (for example 8.4).
        self.legacy_mode = _is_legacy_mode(self.config)

        # Check whether users have disabled datastore_search_sql
        self.enable_sql_search = p.toolkit.asbool(
            self.config.get('ckan.datastore.sqlsearch.enabled', True))

        datapusher_formats = config.get('datapusher.formats', '').split()
        self.datapusher_formats = datapusher_formats or DEFAULT_FORMATS

        # Check whether we are running one of the paster commands which means
        # that we should ignore the following tests.
        if sys.argv[0].split('/')[-1] == 'paster' and 'datastore' in sys.argv[1:]:
            log.warn('Omitting permission checks because you are '
                     'running paster commands.')
            return

        self.ckan_url = self.config['sqlalchemy.url']
        self.write_url = self.config['ckan.datastore.write_url']
        if self.legacy_mode:
            self.read_url = self.write_url
            log.warn('Legacy mode active. '
                     'The sql search will not be available.')
        else:
            self.read_url = self.config['ckan.datastore.read_url']

        self.read_engine = db._get_engine(
            {'connection_url': self.read_url})
        if not model.engine_is_pg(self.read_engine):
            log.warn('We detected that you do not use a PostgreSQL '
                     'database. The DataStore will NOT work and DataStore '
                     'tests will be skipped.')
            return

        if self._is_read_only_database():
            log.warn('We detected that CKAN is running on a read '
                     'only database. Permission checks and the creation '
                     'of _table_metadata are skipped.')
        else:
            self._check_urls_and_permissions()
            self._create_alias_table()

    def notify(self, entity, operation=None):
        if not isinstance(entity, model.Package) or self.legacy_mode:
            return
        # if a resource is new, it cannot have a datastore resource, yet
        if operation == model.domain_object.DomainObjectOperation.changed:
            context = {'model': model, 'ignore_auth': True}
            if entity.private:
                func = p.toolkit.get_action('datastore_make_private')
            else:
                func = p.toolkit.get_action('datastore_make_public')
            for resource in entity.resources:
                try:
                    func(context, {
                        'connection_url': self.write_url,
                        'resource_id': resource.id})
                except p.toolkit.ObjectNotFound:
                    pass

    def _log_or_raise(self, message):
        if self.config.get('debug'):
            log.critical(message)
        else:
            raise DatastoreException(message)

    def _check_urls_and_permissions(self):
        # Make sure that the right permissions are set
        # so that no harmful queries can be made

        if self._same_ckan_and_datastore_db():
            self._log_or_raise('CKAN and DataStore database '
                               'cannot be the same.')

        # in legacy mode, the read and write url are the same (both write url)
        # consequently the same url check and and write privilege check
        # don't make sense
        if not self.legacy_mode:
            if self._same_read_and_write_url():
                self._log_or_raise('The write and read-only database '
                                   'connection urls are the same.')

            if not self._read_connection_has_correct_privileges():
                self._log_or_raise('The read-only user has write privileges.')

    def _is_read_only_database(self):
        ''' Returns True if no connection has CREATE privileges on the public
        schema. This is the case if replication is enabled.'''
        for url in [self.ckan_url, self.write_url, self.read_url]:
            connection = db._get_engine({'connection_url': url}).connect()
            try:
                sql = u"SELECT has_schema_privilege('public', 'CREATE')"
                is_writable = connection.execute(sql).first()[0]
            finally:
                connection.close()
            if is_writable:
                return False
        return True

    def _same_ckan_and_datastore_db(self):
        '''Returns True if the CKAN and DataStore db are the same'''
        return self._get_db_from_url(self.ckan_url) == self._get_db_from_url(self.read_url)

    def _get_db_from_url(self, url):
        db_url = sa_url.make_url(url)
        return db_url.host, db_url.port, db_url.database

    def _same_read_and_write_url(self):
        return self.write_url == self.read_url

    def _read_connection_has_correct_privileges(self):
        ''' Returns True if the right permissions are set for the read
        only user. A table is created by the write user to test the
        read only user.
        '''
        write_connection = db._get_engine(
            {'connection_url': self.write_url}).connect()
        read_connection_user = sa_url.make_url(self.read_url).username

        drop_foo_sql = u'DROP TABLE IF EXISTS _foo'

        write_connection.execute(drop_foo_sql)

        try:
            write_connection.execute(u'CREATE TEMP TABLE _foo ()')
            for privilege in ['INSERT', 'UPDATE', 'DELETE']:
                test_privilege_sql = u"SELECT has_table_privilege(%s, '_foo', %s)"
                have_privilege = write_connection.execute(
                    test_privilege_sql, (read_connection_user, privilege)).first()[0]
                if have_privilege:
                    return False
        finally:
            write_connection.execute(drop_foo_sql)
            write_connection.close()
        return True

    def _create_alias_table(self):
        mapping_sql = '''
            SELECT DISTINCT
                substr(md5(dependee.relname || COALESCE(dependent.relname, '')), 0, 17) AS "_id",
                dependee.relname AS name,
                dependee.oid AS oid,
                dependent.relname AS alias_of
                -- dependent.oid AS oid
            FROM
                pg_class AS dependee
                LEFT OUTER JOIN pg_rewrite AS r ON r.ev_class = dependee.oid
                LEFT OUTER JOIN pg_depend AS d ON d.objid = r.oid
                LEFT OUTER JOIN pg_class AS dependent ON d.refobjid = dependent.oid
            WHERE
                (dependee.oid != dependent.oid OR dependent.oid IS NULL) AND
                (dependee.relname IN (SELECT tablename FROM pg_catalog.pg_tables)
                    OR dependee.relname IN (SELECT viewname FROM pg_catalog.pg_views)) AND
                dependee.relnamespace = (SELECT oid FROM pg_namespace WHERE nspname='public')
            ORDER BY dependee.oid DESC;
        '''
        create_alias_table_sql = u'CREATE OR REPLACE VIEW "_table_metadata" AS {0}'.format(mapping_sql)
        try:
            connection = db._get_engine(
                {'connection_url': self.write_url}).connect()
            connection.execute(create_alias_table_sql)
        finally:
            connection.close()

    def get_actions(self):
        actions = {'datastore_create': action.datastore_create,
                   'datastore_upsert': action.datastore_upsert,
                   'datastore_delete': action.datastore_delete,
                   'datastore_search': action.datastore_search,
                   'datastore_info': action.datastore_info,
                  }
        if not self.legacy_mode:
            if self.enable_sql_search:
                # Only enable search_sql if the config does not disable it
                actions.update({'datastore_search_sql':
                                 action.datastore_search_sql})
            actions.update({
                'datastore_make_private': action.datastore_make_private,
                'datastore_make_public': action.datastore_make_public})
        return actions

    def get_auth_functions(self):
        return {'datastore_create': auth.datastore_create,
                'datastore_upsert': auth.datastore_upsert,
                'datastore_delete': auth.datastore_delete,
                'datastore_info': auth.datastore_info,
                'datastore_search': auth.datastore_search,
                'datastore_search_sql': auth.datastore_search_sql,
                'datastore_change_permissions': auth.datastore_change_permissions}

    def before_map(self, m):
        m.connect('/datastore/dump/{resource_id}',
                  controller='ckanext.datastore.controller:DatastoreController',
                  action='dump')
        return m

    def before_show(self, resource_dict):
        # Modify the resource url of datastore resources so that
        # they link to the datastore dumps.
        if resource_dict.get('url_type') == 'datastore':
            resource_dict['url'] = p.toolkit.url_for(
                controller='ckanext.datastore.controller:DatastoreController',
                action='dump', resource_id=resource_dict['id'],
                qualified=True)

        if 'datastore_active' not in resource_dict:
            resource_dict[u'datastore_active'] = False

        return resource_dict

    def datastore_validate(self, context, data_dict, fields_types):
        column_names = fields_types.keys()
        fields = data_dict.get('fields')
        if fields:
            data_dict['fields'] = list(set(fields) - set(column_names))

        filters = data_dict.get('filters', {})
        for key in filters.keys():
            if key in fields_types:
                del filters[key]

        q = data_dict.get('q')
        if q:
            if isinstance(q, basestring):
                del data_dict['q']
            elif isinstance(q, dict):
                for key in q.keys():
                    if key in fields_types and isinstance(q[key], basestring):
                        del q[key]

        language = data_dict.get('language')
        if language:
            if isinstance(language, basestring):
                del data_dict['language']

        plain = data_dict.get('plain')
        if plain:
            if isinstance(plain, bool):
                del data_dict['plain']

        distinct = data_dict.get('distinct')
        if distinct:
            if isinstance(distinct, bool):
                del data_dict['distinct']

        sort_clauses = data_dict.get('sort')
        if sort_clauses:
            invalid_clauses = [c for c in sort_clauses
                               if not self._parse_sort_clause(c, fields_types)]
            data_dict['sort'] = invalid_clauses

        limit = data_dict.get('limit')
        if limit:
            is_positive_int = datastore_helpers.validate_int(limit,
                                                             non_negative=True)
            is_all = isinstance(limit, basestring) and limit.lower() == 'all'
            if is_positive_int or is_all:
                del data_dict['limit']

        offset = data_dict.get('offset')
        if offset:
            is_positive_int = datastore_helpers.validate_int(offset,
                                                             non_negative=True)
            if is_positive_int:
                del data_dict['offset']

        return data_dict

    def _parse_sort_clause(self, clause, fields_types):
        clause_match = re.match(u'^(.+?)( +(asc|desc) *)?$', clause, re.I)

        if not clause_match:
            return False

        field = clause_match.group(1)
        if field[0] == field[-1] == u'"':
            field = field[1:-1]
        sort = (clause_match.group(3) or u'asc').lower()

        if field not in fields_types:
            return False

        return field, sort

    def datastore_delete(self, context, data_dict, fields_types, query_dict):
        query_dict['where'] += self._where(data_dict, fields_types)
        return query_dict

    def datastore_search(self, context, data_dict, fields_types, query_dict):
        fields = data_dict.get('fields')

        if fields:
            field_ids = datastore_helpers.get_list(fields)
        else:
            field_ids = fields_types.keys()

        ts_query, rank_column = self._textsearch_query(data_dict)
        limit = data_dict.get('limit', 100)
        offset = data_dict.get('offset', 0)

        sort = self._sort(data_dict, fields_types)
        where = self._where(data_dict, fields_types)

        select_cols = [u'"{0}"'.format(field_id) for field_id in field_ids] +\
                      [u'count(*) over() as "_full_count" %s' % rank_column]

        query_dict['distinct'] = data_dict.get('distinct', False)
        query_dict['select'] += select_cols
        query_dict['ts_query'] = ts_query
        query_dict['sort'] += sort
        query_dict['where'] += where
        query_dict['limit'] = limit
        query_dict['offset'] = offset

        return query_dict

    def _where(self, data_dict, fields_types):
        filters = data_dict.get('filters', {})
        clauses = []

        for field, value in filters.iteritems():
            if field not in fields_types:
                continue
            field_array_type = self._is_array_type(fields_types[field])
            if isinstance(value, list) and not field_array_type:
                clause_str = (u'"{0}" in ({1})'.format(field,
                              ','.join(['%s'] * len(value))))
                clause = (clause_str,) + tuple(value)
            else:
                clause = (u'"{0}" = %s'.format(field), value)
            clauses.append(clause)

        # add full-text search where clause
        q = data_dict.get('q')
        if q:
            if isinstance(q, basestring):
                ts_query_alias = self._ts_query_alias()
                clause_str = u'_full_text @@ {0}'.format(ts_query_alias)
                clauses.append((clause_str,))
            elif isinstance(q, dict):
                lang = self._fts_lang(data_dict.get('lang'))
                for field, value in q.iteritems():
                    if field not in fields_types:
                        continue
                    query_field = self._ts_query_alias(field)

                    ftyp = fields_types[field]
                    if not datastore_helpers.should_fts_index_field_type(ftyp):
                        clause_str = u'_full_text @@ {0}'.format(query_field)
                        clauses.append((clause_str,))

                    clause_str = (u'to_tsvector({0}, cast("{1}" as text)) '
                                  u'@@ {2}').format(literal_string(lang),
                                                    field, query_field)
                    clauses.append((clause_str,))

        return clauses

    def _is_array_type(self, field_type):
        return field_type.startswith('_')

    def _sort(self, data_dict, fields_types):
        sort = data_dict.get('sort')
        if not sort:
            q = data_dict.get('q')
            if q:
                if isinstance(q, basestring):
                    return [self._ts_rank_alias()]
                elif isinstance(q, dict):
                    return [self._ts_rank_alias(field) for field in q
                            if field not in fields_types]
            else:
                return []

        clauses = datastore_helpers.get_list(sort, False)

        clause_parsed = []

        for clause in clauses:
            field, sort = self._parse_sort_clause(clause, fields_types)
            clause_parsed.append(
                u'{0} {1}'.format(datastore_helpers.identifier(field), sort))

        return clause_parsed

    def _textsearch_query(self, data_dict):
        q = data_dict.get('q')
        lang = self._fts_lang(data_dict.get('lang'))

        if not q:
            return '', ''

        statements = []
        rank_columns = []
        plain = data_dict.get('plain', True)
        if isinstance(q, basestring):
            query, rank = self._build_query_and_rank_statements(lang,
                                                                q,
                                                                plain)
            statements.append(query)
            rank_columns.append(rank)
        elif isinstance(q, dict):
            for field, value in q.iteritems():
                query, rank = self._build_query_and_rank_statements(lang,
                                                                    value,
                                                                    plain,
                                                                    field)
                statements.append(query)
                rank_columns.append(rank)

        statements_str = ', ' + ', '.join(statements)
        rank_columns_str = ', ' + ', '.join(rank_columns)
        return statements_str, rank_columns_str

    def _fts_lang(self, lang=None):
        default_fts_lang = config.get('ckan.datastore.default_fts_lang')
        if default_fts_lang is None:
            default_fts_lang = u'english'
        return lang or default_fts_lang

    def _build_query_and_rank_statements(self, lang, query, plain, field=None):
        query_alias = self._ts_query_alias(field)
        rank_alias = self._ts_rank_alias(field)
        lang_literal = literal_string(lang)
        query_literal = literal_string(query)
        if plain:
            statement = u"plainto_tsquery({lang_literal}, {query_literal}) {query_alias}"
        else:
            statement = u"to_tsquery({lang_literal}, {query_literal}) {query_alias}"
        statement = statement.format(lang_literal=lang_literal,
            query_literal=query_literal, query_alias=query_alias)
        if field is None:
            rank_field = '_full_text'
        else:
            rank_field = u'to_tsvector({lang_literal}, cast("{field}" as text))'
            rank_field = rank_field.format(lang_literal=lang_literal, field=field)
        rank_statement = u'ts_rank({rank_field}, {query_alias}, 32) AS {alias}'
        rank_statement = rank_statement.format(rank_field=rank_field,
                                               query_alias=query_alias,
                                               alias=rank_alias)
        return statement, rank_statement

    def _ts_query_alias(self, field=None):
        query_alias = u'query'
        if field:
            query_alias += u' ' + field
        return u'"{0}"'.format(query_alias)

    def _ts_rank_alias(self, field=None):
        rank_alias = u'rank'
        if field:
            rank_alias += u' ' + field
        return u'"{0}"'.format(rank_alias)



# encoding: utf-8

import mock
import wsgiref
from nose.tools import assert_equals, assert_not_equals, eq_
from routes import url_for

import ckan.plugins as p
import ckan.tests.helpers as helpers

from ckan.config.middleware import AskAppDispatcherMiddleware
from ckan.config.middleware.flask_app import CKANFlask
from ckan.config.middleware.pylons_app import CKANPylonsApp


class TestPylonsResponseCleanupMiddleware(helpers.FunctionalTestBase):
    @classmethod
    def _apply_config_changes(cls, config):
        config['ckan.use_pylons_response_cleanup_middleware'] = True

    def test_homepage_with_middleware_activated(self):
        '''Test the home page renders with the middleware activated

        We are just testing the home page renders without any troubles and that
        the middleware has not done anything strange to the response string'''
        app = self._get_test_app()
        response = app.get(url=url_for(controller='home', action='index'))

        assert_equals(200, response.status_int)
        # make sure we haven't overwritten the response too early.
        assert_not_equals(
            'response cleared by pylons response cleanup middleware',
            response.body
        )


class TestAppDispatcherPlain(object):
    '''
    These tests need the test app to be created at specific times to not affect
    the mocks, so they don't extend FunctionalTestBase
    '''

    def test_flask_can_handle_request_is_called_with_environ(self):

        with mock.patch.object(CKANFlask, 'can_handle_request') as \
                mock_can_handle_request:
            # We need set this otherwise the mock object is returned
            mock_can_handle_request.return_value = (False, 'flask_app')

            app = helpers._get_test_app()
            # We want our CKAN app, not the WebTest one
            ckan_app = app.app

            environ = {
                'PATH_INFO': '/',
            }
            wsgiref.util.setup_testing_defaults(environ)
            start_response = mock.MagicMock()

            ckan_app(environ, start_response)

            assert mock_can_handle_request.called_with(environ)

    def test_pylons_can_handle_request_is_called_with_environ(self):

        with mock.patch.object(CKANPylonsApp, 'can_handle_request') as \
                mock_can_handle_request:

            # We need set this otherwise the mock object is returned
            mock_can_handle_request.return_value = (True, 'pylons_app', 'core')

            app = helpers._get_test_app()
            # We want our CKAN app, not the WebTest one
            ckan_app = app.app

            environ = {
                'PATH_INFO': '/',
            }
            wsgiref.util.setup_testing_defaults(environ)
            start_response = mock.MagicMock()

            ckan_app(environ, start_response)

            assert mock_can_handle_request.called_with(environ)


class TestAppDispatcher(helpers.FunctionalTestBase):

    @classmethod
    def setup_class(cls):

        super(TestAppDispatcher, cls).setup_class()

        # Add a custom route to the Flask app
        app = cls._get_test_app()

        flask_app = app.app.apps['flask_app']

        def test_view():
            return 'This was served from Flask'

        # This endpoint is defined both in Flask and in Pylons core
        flask_app.add_url_rule('/about', view_func=test_view)

        # This endpoint is defined both in Flask and a Pylons extension
        flask_app.add_url_rule('/pylons_and_flask', view_func=test_view)

    def test_ask_around_is_called(self):

        app = self._get_test_app()
        with mock.patch.object(AskAppDispatcherMiddleware, 'ask_around') as \
                mock_ask_around:
            app.get('/')

            assert mock_ask_around.called

    def test_ask_around_is_called_with_args(self):

        app = self._get_test_app()
        ckan_app = app.app

        environ = {}
        start_response = mock.MagicMock()
        wsgiref.util.setup_testing_defaults(environ)

        with mock.patch.object(AskAppDispatcherMiddleware, 'ask_around') as \
                mock_ask_around:

            ckan_app(environ, start_response)
            assert mock_ask_around.called
            mock_ask_around.assert_called_with(environ)

    def test_ask_around_flask_core_route_get(self):

        app = self._get_test_app()

        # We want our CKAN app, not the WebTest one
        app = app.app

        environ = {
            'PATH_INFO': '/hello',
            'REQUEST_METHOD': 'GET',
        }
        wsgiref.util.setup_testing_defaults(environ)

        answers = app.ask_around(environ)

        # Even though this route is defined in Flask, there is catch all route
        # in Pylons for all requests to point arbitrary urls to templates with
        # the same name, so we get two positive answers
        eq_(answers, [(True, 'flask_app'), (True, 'pylons_app', 'core')])
        # TODO: check Flask origin (core/extension) when that is in place
        # (also on the following tests)

    def test_ask_around_flask_core_route_post(self):

        app = self._get_test_app()

        # We want our CKAN app, not the WebTest one
        app = app.app

        environ = {
            'PATH_INFO': '/hello',
            'REQUEST_METHOD': 'POST',
        }
        wsgiref.util.setup_testing_defaults(environ)

        answers = app.ask_around(environ)

        # Even though this route is defined in Flask, there is catch all route
        # in Pylons for all requests to point arbitrary urls to templates with
        # the same name, so we get two positive answers
        eq_(answers, [(True, 'flask_app'), (True, 'pylons_app', 'core')])

    def test_ask_around_pylons_core_route_get(self):

        app = self._get_test_app()

        # We want our CKAN app, not the WebTest one
        app = app.app

        environ = {
            'PATH_INFO': '/dataset',
            'REQUEST_METHOD': 'GET',
        }
        wsgiref.util.setup_testing_defaults(environ)

        answers = app.ask_around(environ)

        eq_(answers, [(False, 'flask_app'), (True, 'pylons_app', 'core')])

    def test_ask_around_pylons_core_route_post(self):

        app = self._get_test_app()

        # We want our CKAN app, not the WebTest one
        app = app.app

        environ = {
            'PATH_INFO': '/dataset/new',
            'REQUEST_METHOD': 'POST',
        }
        wsgiref.util.setup_testing_defaults(environ)

        answers = app.ask_around(environ)

        eq_(answers, [(False, 'flask_app'), (True, 'pylons_app', 'core')])

    def test_ask_around_pylons_extension_route_get_before_map(self):

        if not p.plugin_loaded('test_routing_plugin'):
            p.load('test_routing_plugin')

        app = self._get_test_app()

        # We want our CKAN app, not the WebTest one
        app = app.app

        environ = {
            'PATH_INFO': '/from_pylons_extension_before_map',
            'REQUEST_METHOD': 'GET',
        }
        wsgiref.util.setup_testing_defaults(environ)

        answers = app.ask_around(environ)

        eq_(answers, [(False, 'flask_app'), (True, 'pylons_app', 'extension')])

        p.unload('test_routing_plugin')

    def test_ask_around_pylons_extension_route_post(self):

        if not p.plugin_loaded('test_routing_plugin'):
            p.load('test_routing_plugin')

        app = self._get_test_app()

        # We want our CKAN app, not the WebTest one
        app = app.app

        environ = {
            'PATH_INFO': '/from_pylons_extension_before_map_post_only',
            'REQUEST_METHOD': 'POST',
        }
        wsgiref.util.setup_testing_defaults(environ)

        answers = app.ask_around(environ)

        eq_(answers, [(False, 'flask_app'), (True, 'pylons_app', 'extension')])

        p.unload('test_routing_plugin')

    def test_ask_around_pylons_extension_route_post_using_get(self):

        if not p.plugin_loaded('test_routing_plugin'):
            p.load('test_routing_plugin')

        app = self._get_test_app()

        # We want our CKAN app, not the WebTest one
        app = app.app

        environ = {
            'PATH_INFO': '/from_pylons_extension_before_map_post_only',
            'REQUEST_METHOD': 'GET',
        }
        wsgiref.util.setup_testing_defaults(environ)

        answers = app.ask_around(environ)

        # We are going to get an answer from Pylons, but just because it will
        # match the catch-all template route, hence the `core` origin.
        eq_(answers, [(False, 'flask_app'), (True, 'pylons_app', 'core')])

        p.unload('test_routing_plugin')

    def test_ask_around_pylons_extension_route_get_after_map(self):

        if not p.plugin_loaded('test_routing_plugin'):
            p.load('test_routing_plugin')

        app = self._get_test_app()

        # We want our CKAN app, not the WebTest one
        app = app.app

        environ = {
            'PATH_INFO': '/from_pylons_extension_after_map',
            'REQUEST_METHOD': 'GET',
        }
        wsgiref.util.setup_testing_defaults(environ)

        answers = app.ask_around(environ)

        eq_(answers, [(False, 'flask_app'), (True, 'pylons_app', 'extension')])

        p.unload('test_routing_plugin')

    def test_ask_around_flask_core_and_pylons_extension_route(self):

        if not p.plugin_loaded('test_routing_plugin'):
            p.load('test_routing_plugin')

        app = self._get_test_app()

        # We want our CKAN app, not the WebTest one
        app = app.app

        environ = {
            'PATH_INFO': '/pylons_and_flask',
            'REQUEST_METHOD': 'GET',
        }
        wsgiref.util.setup_testing_defaults(environ)

        answers = app.ask_around(environ)
        answers = sorted(answers, key=lambda a: a[1])

        eq_(answers, [(True, 'flask_app'), (True, 'pylons_app', 'extension')])

        p.unload('test_routing_plugin')

    def test_flask_core_route_is_served_by_flask(self):

        app = self._get_test_app()

        res = app.get('/hello')

        eq_(res.environ['ckan.app'], 'flask_app')

    # TODO: test flask extension route

    def test_pylons_core_route_is_served_by_pylons(self):

        app = self._get_test_app()

        res = app.get('/dataset')

        eq_(res.environ['ckan.app'], 'pylons_app')

    def test_pylons_extension_route_is_served_by_pylons(self):

        if not p.plugin_loaded('test_routing_plugin'):
            p.load('test_routing_plugin')

        app = self._get_test_app()

        res = app.get('/from_pylons_extension_before_map')

        eq_(res.environ['ckan.app'], 'pylons_app')
        eq_(res.body, 'Hello World, this is served from a Pylons extension')

        p.unload('test_routing_plugin')

    def test_flask_core_and_pylons_extension_route_is_served_by_pylons(self):

        if not p.plugin_loaded('test_routing_plugin'):
            p.load('test_routing_plugin')

        app = self._get_test_app()

        res = app.get('/pylons_and_flask')

        eq_(res.environ['ckan.app'], 'pylons_app')
        eq_(res.body, 'Hello World, this is served from a Pylons extension')

        p.unload('test_routing_plugin')

    def test_flask_core_and_pylons_core_route_is_served_by_flask(self):
        '''
        This should never happen in core, but just in case
        '''
        app = self._get_test_app()

        res = app.get('/about')

        eq_(res.environ['ckan.app'], 'flask_app')
        eq_(res.body, 'This was served from Flask')


class MockRoutingPlugin(p.SingletonPlugin):

    p.implements(p.IRoutes)

    controller = 'ckan.tests.config.test_middleware:MockPylonsController'

    def before_map(self, _map):

        _map.connect('/from_pylons_extension_before_map',
                     controller=self.controller, action='view')

        _map.connect('/from_pylons_extension_before_map_post_only',
                     controller=self.controller, action='view',
                     conditions={'method': 'POST'})
        # This one conflicts with a core Flask route
        _map.connect('/pylons_and_flask',
                     controller=self.controller, action='view')

        return _map

    def after_map(self, _map):

        _map.connect('/from_pylons_extension_after_map',
                     controller=self.controller, action='view')

        return _map


class MockPylonsController(p.toolkit.BaseController):

    def view(self):
        return 'Hello World, this is served from a Pylons extension'



# encoding: utf-8

import os

from babel import Locale
from babel.core import (LOCALE_ALIASES,
                        get_locale_identifier,
                        UnknownLocaleError)
from babel.support import Translations
from paste.deploy.converters import aslist
from pylons import i18n
import pylons


from ckan.common import config
import ckan.i18n
from ckan.plugins import PluginImplementations
from ckan.plugins.interfaces import ITranslation

# Default Portuguese language to Brazilian territory, since
# we don't have a Portuguese territory translation currently.
LOCALE_ALIASES['pt'] = 'pt_BR'


def get_locales_from_config():
    ''' despite the name of this function it gets the locales defined by
    the config AND also the locals available subject to the config. '''
    locales_offered = config.get('ckan.locales_offered', '').split()
    filtered_out = config.get('ckan.locales_filtered_out', '').split()
    locale_default = [config.get('ckan.locale_default', 'en')]
    locale_order = config.get('ckan.locale_order', '').split()

    known_locales = get_locales()
    all_locales = (set(known_locales) |
                   set(locales_offered) |
                   set(locale_order) |
                   set(locale_default))
    all_locales -= set(filtered_out)
    return all_locales


def _get_locales():
    # FIXME this wants cleaning up and merging with get_locales_from_config()
    assert not config.get('lang'), \
        ('"lang" config option not supported - please use ckan.locale_default '
         'instead.')
    locales_offered = config.get('ckan.locales_offered', '').split()
    filtered_out = config.get('ckan.locales_filtered_out', '').split()
    locale_default = config.get('ckan.locale_default', 'en')
    locale_order = config.get('ckan.locale_order', '').split()

    locales = ['en']
    if config.get('ckan.i18n_directory'):
        i18n_path = os.path.join(config.get('ckan.i18n_directory'), 'i18n')
    else:
        i18n_path = os.path.dirname(ckan.i18n.__file__)

    # For every file in the ckan i18n directory see if babel can understand
    # the locale. If yes, add it to the available locales
    for locale in os.listdir(i18n_path):
        try:
            Locale.parse(locale)
            locales.append(locale)
        except (ValueError, UnknownLocaleError):
            # Babel does not know how to make a locale out of this.
            # This is fine since we are passing all files in the
            # ckan.i18n_directory here which e.g. includes the __init__.py
            pass

    assert locale_default in locales, \
        'default language "%s" not available' % locale_default

    locale_list = []
    for locale in locales:
        # no duplicates
        if locale in locale_list:
            continue
        # if offered locales then check locale is offered
        if locales_offered and locale not in locales_offered:
            continue
        # remove if filtered out
        if locale in filtered_out:
            continue
        # ignore the default as it will be added first
        if locale == locale_default:
            continue
        locale_list.append(locale)
    # order the list if specified
    ordered_list = [locale_default]
    for locale in locale_order:
        if locale in locale_list:
            ordered_list.append(locale)
            # added so remove from our list
            locale_list.remove(locale)
    # add any remaining locales not ordered
    ordered_list += locale_list

    return ordered_list

available_locales = None
locales = None
locales_dict = None
_non_translated_locals = None


def get_locales():
    ''' Get list of available locales
    e.g. [ 'en', 'de', ... ]
    '''
    global locales
    if not locales:
        locales = _get_locales()
    return locales


def non_translated_locals():
    ''' These are the locales that are available but for which there are
    no translations. returns a list like ['en', 'de', ...] '''
    global _non_translated_locals
    if not _non_translated_locals:
        locales = config.get('ckan.locale_order', '').split()
        _non_translated_locals = [x for x in locales if x not in get_locales()]
    return _non_translated_locals


def get_locales_dict():
    ''' Get a dict of the available locales
    e.g.  { 'en' : Locale('en'), 'de' : Locale('de'), ... } '''
    global locales_dict
    if not locales_dict:
        locales = _get_locales()
        locales_dict = {}
        for locale in locales:
            locales_dict[str(locale)] = Locale.parse(locale)
    return locales_dict


def get_available_locales():
    ''' Get a list of the available locales
    e.g.  [ Locale('en'), Locale('de'), ... ] '''
    global available_locales
    if not available_locales:
        available_locales = []
        for locale in get_locales():
            # Add the short names for the locales. This equals the filename
            # of the ckan translation files as opposed to the long name
            # that includes the script which is generated by babel
            # so e.g. `zn_CH` instead of `zn_Hans_CH` this is needed
            # to properly construct urls with url_for
            parsed_locale = Locale.parse(locale)
            parsed_locale.short_name = locale

            # Add the full identifier (eg `pt_BR`) to the locale classes,
            # as it does not offer a way of accessing it directly
            parsed_locale.identifier = \
                get_identifier_from_locale_class(parsed_locale)
            available_locales.append(parsed_locale)
    return available_locales


def get_identifier_from_locale_class(locale):
    return get_locale_identifier(
        (locale.language,
         locale.territory,
         locale.script,
         locale.variant))


def _set_lang(lang):
    ''' Allows a custom i18n directory to be specified.
    Creates a fake config file to pass to pylons.i18n.set_lang, which
    sets the Pylons root path to desired i18n_directory.
    This is needed as Pylons will only look for an i18n directory in
    the application root.'''
    if config.get('ckan.i18n_directory'):
        fake_config = {'pylons.paths': {'root': config['ckan.i18n_directory']},
                       'pylons.package': config['pylons.package']}
        i18n.set_lang(lang, pylons_config=fake_config, class_=Translations)
    else:
        i18n.set_lang(lang, class_=Translations)


def handle_request(request, tmpl_context):
    ''' Set the language for the request '''
    lang = request.environ.get('CKAN_LANG') or \
        config.get('ckan.locale_default', 'en')
    if lang != 'en':
        set_lang(lang)

    for plugin in PluginImplementations(ITranslation):
        if lang in plugin.i18n_locales():
            _add_extra_translations(plugin.i18n_directory(), lang,
                                    plugin.i18n_domain())

    extra_directory = config.get('ckan.i18n.extra_directory')
    extra_domain = config.get('ckan.i18n.extra_gettext_domain')
    extra_locales = aslist(config.get('ckan.i18n.extra_locales'))
    if extra_directory and extra_domain and extra_locales:
        if lang in extra_locales:
            _add_extra_translations(extra_directory, lang, extra_domain)

    tmpl_context.language = lang
    return lang


def _add_extra_translations(dirname, locales, domain):
    translator = Translations.load(dirname=dirname, locales=locales,
                                   domain=domain)
    try:
        pylons.translator.merge(translator)
    except AttributeError:
        # this occurs when an extension has 'en' translations that
        # replace the default strings. As set_lang has not been run,
        # pylons.translation is the NullTranslation, so we have to
        # replace the StackedObjectProxy ourselves manually.
        environ = pylons.request.environ
        environ['pylons.pylons'].translator = translator
        if 'paste.registry' in environ:
            environ['paste.registry'].replace(pylons.translator,
                                              translator)


def get_lang():
    ''' Returns the current language. Based on babel.i18n.get_lang but
    works when set_lang has not been run (i.e. still in English). '''
    langs = i18n.get_lang()
    if langs:
        return langs[0]
    else:
        return 'en'


def set_lang(language_code):
    ''' Wrapper to pylons call '''
    if language_code in non_translated_locals():
        language_code = config.get('ckan.locale_default', 'en')
    if language_code != 'en':
        _set_lang(language_code)



# encoding: utf-8

import logging

import ckan.plugins as p
import ckan.lib.base as base
import ckan.lib.helpers as core_helpers
import ckanext.datapusher.logic.action as action
import ckanext.datapusher.logic.auth as auth
import ckanext.datapusher.helpers as helpers
import ckan.logic as logic
import ckan.model as model
import ckan.plugins.toolkit as toolkit

from ckan.common import _

log = logging.getLogger(__name__)
_get_or_bust = logic.get_or_bust

DEFAULT_FORMATS = [
    'csv', 'xls', 'xlsx', 'tsv', 'application/csv',
    'application/vnd.ms-excel',
    'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
    'ods', 'application/vnd.oasis.opendocument.spreadsheet',
]


class DatastoreException(Exception):
    pass


class ResourceDataController(base.BaseController):

    def resource_data(self, id, resource_id):

        if toolkit.request.method == 'POST':
            try:
                toolkit.c.pkg_dict = p.toolkit.get_action('datapusher_submit')(
                    None, {'resource_id': resource_id}
                )
            except logic.ValidationError:
                pass

            base.redirect(core_helpers.url_for(
                controller='ckanext.datapusher.plugin:ResourceDataController',
                action='resource_data',
                id=id,
                resource_id=resource_id)
            )

        try:
            toolkit.c.pkg_dict = p.toolkit.get_action('package_show')(
                None, {'id': id}
            )
            toolkit.c.resource = p.toolkit.get_action('resource_show')(
                None, {'id': resource_id}
            )
        except (logic.NotFound, logic.NotAuthorized):
            base.abort(404, _('Resource not found'))

        try:
            datapusher_status = p.toolkit.get_action('datapusher_status')(
                None, {'resource_id': resource_id}
            )
        except logic.NotFound:
            datapusher_status = {}
        except logic.NotAuthorized:
            base.abort(403, _('Not authorized to see this page'))

        return base.render('package/resource_data.html',
                           extra_vars={'status': datapusher_status})


class DatapusherPlugin(p.SingletonPlugin):
    p.implements(p.IConfigurable, inherit=True)
    p.implements(p.IActions)
    p.implements(p.IAuthFunctions)
    p.implements(p.IResourceUrlChange)
    p.implements(p.IDomainObjectModification, inherit=True)
    p.implements(p.ITemplateHelpers)
    p.implements(p.IRoutes, inherit=True)

    legacy_mode = False
    resource_show_action = None

    def configure(self, config):
        self.config = config

        datapusher_formats = config.get('ckan.datapusher.formats', '').lower()
        self.datapusher_formats = datapusher_formats.split() or DEFAULT_FORMATS

        for config_option in ('ckan.site_url', 'ckan.datapusher.url',):
            if not config.get(config_option):
                raise Exception(
                    'Config option `{0}` must be set to use the DataPusher.'
                    .format(config_option))

    def notify(self, entity, operation=None):
        if isinstance(entity, model.Resource):
            if (operation == model.domain_object.DomainObjectOperation.new or
                    not operation):
                # if operation is None, resource URL has been changed, as
                # the notify function in IResourceUrlChange only takes
                # 1 parameter
                context = {'model': model, 'ignore_auth': True,
                           'defer_commit': True}
                if (entity.format and
                        entity.format.lower() in self.datapusher_formats and
                        entity.url_type != 'datapusher'):

                    try:
                        task = p.toolkit.get_action('task_status_show')(
                            context, {
                                'entity_id': entity.id,
                                'task_type': 'datapusher',
                                'key': 'datapusher'}
                        )
                        if task.get('state') == 'pending':
                            # There already is a pending DataPusher submission,
                            # skip this one ...
                            log.debug(
                                'Skipping DataPusher submission for '
                                'resource {0}'.format(entity.id))
                            return
                    except p.toolkit.ObjectNotFound:
                        pass

                    try:
                        log.debug('Submitting resource {0}'.format(entity.id) +
                                  ' to DataPusher')
                        p.toolkit.get_action('datapusher_submit')(context, {
                            'resource_id': entity.id
                        })
                    except p.toolkit.ValidationError, e:
                        # If datapusher is offline want to catch error instead
                        # of raising otherwise resource save will fail with 500
                        log.critical(e)
                        pass

    def before_map(self, m):
        m.connect(
            'resource_data', '/dataset/{id}/resource_data/{resource_id}',
            controller='ckanext.datapusher.plugin:ResourceDataController',
            action='resource_data', ckan_icon='cloud-upload')
        return m

    def get_actions(self):
        return {'datapusher_submit': action.datapusher_submit,
                'datapusher_hook': action.datapusher_hook,
                'datapusher_status': action.datapusher_status}

    def get_auth_functions(self):
        return {'datapusher_submit': auth.datapusher_submit,
                'datapusher_status': auth.datapusher_status}

    def get_helpers(self):
        return {
            'datapusher_status': helpers.datapusher_status,
            'datapusher_status_description':
            helpers.datapusher_status_description,
        }



# encoding: utf-8

import ckan.plugins as plugins


def group_create(context, data_dict=None):
    return {'success': False, 'msg': 'No one is allowed to create groups'}


class ExampleIAuthFunctionsPlugin(plugins.SingletonPlugin):
    plugins.implements(plugins.IAuthFunctions)

    def get_auth_functions(self):
        return {'group_create': group_create}



# encoding: utf-8

import ckan.model as model

Activity = model.Activity
ActivityDetail = model.ActivityDetail


class TestActivityDetail(object):
    def test_by_activity_id(self):
        activity = Activity('user-id', 'object-id',
                            'revision-id', 'activity-type')
        activity.save()
        activity_detail = ActivityDetail(activity.id, 'object-id',
                                         'object-type', 'activity-type')
        activity_detail.save()
        activities = ActivityDetail.by_activity_id(activity.id)
        assert activities == [activity_detail], activity_detail



# encoding: utf-8

from nose.tools import assert_equal 

from ckan import model
from ckan.lib.create_test_data import CreateTestData

from ckan.tests.legacy.functional.api.base import BaseModelApiTestCase
from ckan.tests.legacy.functional.api.base import Api1TestCase as Version1TestCase
from ckan.tests.legacy.functional.api.base import Api2TestCase as Version2TestCase

class LicensesTestCase(BaseModelApiTestCase):

    @classmethod
    def setup_class(cls):
        CreateTestData.create()

    @classmethod
    def teardown_class(cls):
        model.repo.rebuild_db()

    def test_register_get_ok(self):
        from ckan.model.license import LicenseRegister
        register = LicenseRegister()
        assert len(register), "No changesets found in model."
        offset = self.offset('/rest/licenses')
        res = self.app.get(offset, status=[200])
        licenses_data = self.data_from_res(res)
        assert len(licenses_data) == len(register), (len(licenses_data), len(register))
        for license_data in licenses_data:
            id = license_data['id']
            license = register[id]
            assert license['title'] == license.title
            assert license['url'] == license.url


class TestLicensesVersion1(Version1TestCase, LicensesTestCase): pass
class TestLicensesVersion2(Version2TestCase, LicensesTestCase): pass



# encoding: utf-8

from ckan.tests.legacy import *

import ckan.model as model

class TestRevisionPurge:
    
    @classmethod
    def setup_class(self):
        model.Session.remove()
        CreateTestData.create()

    @classmethod
    def teardown_class(self):
        model.repo.rebuild_db()

    def setup(self):
        self.pkgname = u'revision-purge-test'

        model.repo.new_revision()
        self.pkg = model.Package(name=self.pkgname)
        self.old_url = u'abc.com'
        self.pkg.url = self.old_url
        tag1 = model.Tag.by_name(u'russian')
        tag2 = model.Tag.by_name(u'tolstoy')
        self.pkg.add_tag(tag1)
        self.pkg.add_tag(tag2)
        model.repo.commit_and_remove()

        txn2 = model.repo.new_revision()
        pkg = model.Package.by_name(self.pkgname)
        newurl = u'blah.com'
        pkg.url = newurl
        for tag in pkg.get_tags():
            pkg.remove_tag(tag)
        self.pkgname2 = u'revision-purge-test-2'
        self.pkg_new = model.Package(name=self.pkgname2)
        model.repo.commit_and_remove()

    def teardown(self):
        model.Session.remove()
        pkg_new = model.Package.by_name(self.pkgname2)
        if pkg_new:
            pkg_new.purge()
        pkg = model.Package.by_name(self.pkgname)
        pkg.purge()
        model.Session.commit()
        model.Session.remove()

    def test_1(self):
        rev = model.repo.youngest_revision()
        model.repo.purge_revision(rev, leave_record=True)

        rev = model.repo.youngest_revision()
        pkg = model.Package.by_name(self.pkgname)

        assert rev.message.startswith('PURGED'), rev.message
        assert pkg.url == self.old_url
        pkg2 = model.Package.by_name(self.pkgname2)
        assert pkg2 is None, 'pkgname2 should no longer exist'
        assert len(pkg.get_tags()) == 2

    def test_2(self):
        rev = model.repo.youngest_revision()
        num = rev.id
        model.repo.purge_revision(rev, leave_record=True)

        rev = model.repo.youngest_revision()
        # TODO: should youngest_revision be made purge aware
        # (requires state on revision)
        assert rev.id == num

    def test_purge_first_revision(self):
        rev = model.repo.youngest_revision()
        num = rev.id
        q = model.repo.history()
        q = q.order_by(model.Revision.timestamp.desc())
        q = q.limit(2)
        rev2 = q.all()[1]
        model.repo.purge_revision(rev, leave_record=True)

        rev = model.repo.youngest_revision()
        assert rev.id == num
        # either none or should equal num - 2 or be None (if no lower revision)
        pkg = model.Package.by_name(self.pkgname)
        assert len(pkg.all_revisions) == 1




# encoding: utf-8

from collections import defaultdict

import ckan.plugins as p
import ckan.tests.legacy.mock_plugin as mock_plugin


class MapperPlugin(p.SingletonPlugin):
    p.implements(p.IMapper, inherit=True)

    def __init__(self, *args, **kw):
        self.calls = []

    def before_insert(self, mapper, conn, instance):
        self.calls.append(('before_insert', instance.name))

    def after_insert(self, mapper, conn, instance):
        self.calls.append(('after_insert', instance.name))

    def before_delete(self, mapper, conn, instance):
        self.calls.append(('before_delete', instance.name))

    def after_delete(self, mapper, conn, instance):
        self.calls.append(('after_delete', instance.name))


class MapperPlugin2(MapperPlugin):
    p.implements(p.IMapper)


class SessionPlugin(p.SingletonPlugin):
    p.implements(p.ISession, inherit=True)

    def __init__(self, *args, **kw):
        self.added = []
        self.deleted = []

    def before_insert(self, mapper, conn, instance):
        self.added.append(instance)

    def before_delete(self, mapper, conn, instance):
        self.deleted.append(instance)

class RoutesPlugin(p.SingletonPlugin):
    p.implements(p.IRoutes, inherit=True)

    def __init__(self, *args, **kw):
        self.calls_made = []

    def before_map(self, map):
        self.calls_made.append('before_map')
        return map

    def after_map(self, map):
        self.calls_made.append('after_map')
        return map


class PluginObserverPlugin(mock_plugin.MockSingletonPlugin):
    p.implements(p.IPluginObserver)

class ActionPlugin(p.SingletonPlugin):
    p.implements(p.IActions)

    def get_actions(self):
        return {'status_show': lambda context, data_dict: {}}

class AuthPlugin(p.SingletonPlugin):
    p.implements(p.IAuthFunctions)

    def get_auth_functions(self):
        return {'package_list': lambda context, data_dict: {}}

class MockGroupControllerPlugin(p.SingletonPlugin):
    p.implements(p.IGroupController)

    def __init__(self, *args, **kw):
        self.calls = defaultdict(int)

    def read(self, entity):
        self.calls['read'] += 1

    def create(self, entity):
        self.calls['create'] += 1

    def edit(self, entity):
        self.calls['edit'] += 1

    def authz_add_role(self, object_role):
        self.calls['authz_add_role'] += 1

    def authz_remove_role(self, object_role):
        self.calls['authz_remove_role'] += 1

    def delete(self, entity):
        self.calls['delete'] += 1

    def before_view(self, data_dict):
        self.calls['before_view'] += 1
        return data_dict


class MockPackageControllerPlugin(p.SingletonPlugin):
    p.implements(p.IPackageController)

    def __init__(self, *args, **kw):
        self.calls = defaultdict(int)

    def read(self, entity):
        self.calls['read'] += 1

    def create(self, entity):
        self.calls['create'] += 1

    def edit(self, entity):
        self.calls['edit'] += 1

    def authz_add_role(self, object_role):
        self.calls['authz_add_role'] += 1

    def authz_remove_role(self, object_role):
        self.calls['authz_remove_role'] += 1

    def delete(self, entity):
        self.calls['delete'] += 1

    def before_search(self, search_params):
        self.calls['before_search'] += 1
        return search_params

    def after_search(self, search_results, search_params):
        self.calls['after_search'] += 1
        return search_results

    def before_index(self, data_dict):
        self.calls['before_index'] += 1
        return data_dict

    def before_view(self, data_dict):
        self.calls['before_view'] += 1
        return data_dict

    def after_create(self, context, data_dict):
        self.calls['after_create'] += 1
        self.id_in_dict = 'id' in data_dict

        return data_dict

    def after_update(self, context, data_dict):
        self.calls['after_update'] += 1
        return data_dict

    def after_delete(self, context, data_dict):
        self.calls['after_delete'] += 1
        return data_dict

    def after_show(self, context, data_dict):
        self.calls['after_show'] += 1
        return data_dict

    def update_facet_titles(self, facet_titles):
        return facet_titles



class MockResourcePreviewExtension(mock_plugin.MockSingletonPlugin):
    p.implements(p.IResourcePreview)

    def __init__(self, *args, **kw):
        self.calls = defaultdict(int)

    def setup_template_variables(self, context, data_dict):
        self.calls['setup_template_variables'] += 1

    def can_preview(self, data_dict):
        assert(isinstance(data_dict['resource'], dict))
        assert(isinstance(data_dict['package'], dict))
        assert('on_same_domain' in data_dict['resource'])

        self.calls['can_preview'] += 1
        return data_dict['resource']['format'].lower() == 'mock'

    def preview_template(self, context, data_dict):
        assert(isinstance(data_dict['resource'], dict))
        assert(isinstance(data_dict['package'], dict))

        self.calls['preview_templates'] += 1
        return 'tests/mock_resource_preview_template.html'


class JsonMockResourcePreviewExtension(mock_plugin.MockSingletonPlugin):
    p.implements(p.IResourcePreview)

    def __init__(self, *args, **kw):
        self.calls = defaultdict(int)

    def setup_template_variables(self, context, data_dict):
        self.calls['setup_template_variables'] += 1

    def can_preview(self, data_dict):
        self.calls['can_preview'] += 1
        return data_dict['resource']['format'].lower() == 'json'

    def preview_template(self, context, data_dict):
        self.calls['preview_templates'] += 1
        return 'tests/mock_json_resource_preview_template.html'


# importing this file loads all these extensions by default
# so clean up the extensions
p.plugins_update()



# encoding: utf-8

import unittest
import nose.tools
import pyutilib.component.core

import ckan.plugins
#import ckanext.datastore.plugin as plugin


class _TestConfiguration(unittest.TestCase):
    # FIXME This entire test class is broken and currently disabled.  A test
    # should not be changing the plugin itself WTF!  I'm not sure if these
    # tests have any value whatsoever.  Anyhow the plugin is not capable of
    # being so tested.  Also why do these test raise a custom exception?
    def setUp(self):
        self._original_plugin = ckan.plugins.unload('datastore')
        pyutilib.component.core.PluginGlobals.singleton_services()[plugin.DatastorePlugin] = True
        self.p = pyutilib.component.core.PluginGlobals.singleton_services()[plugin.DatastorePlugin] = ckan.plugins.load('datastore')

    def tearDown(self):
        ckan.plugins.unload('datastore')
        pyutilib.component.core.PluginGlobals.singleton_services()[plugin.DatastorePlugin] = self._original_plugin

    def test_set_legacy_mode(self):
        c = {
            'sqlalchemy.url': 'bar',
            'ckan.datastore.write_url': 'foo'
        }
        try:
            self.p.configure(c)
        except Exception:
            pass
        assert self.p.legacy_mode
        assert self.p.write_url == 'foo'
        assert self.p.read_url == 'foo'

    def test_check_separate_write_and_read_url(self):
        self.p.write_url = 'postgresql://u:pass@localhost/ds'
        self.p.read_url = 'postgresql://u:pass@localhost/ds'
        assert self.p._same_read_and_write_url()

        self.p.write_url = 'postgresql://u:pass@localhost/ds'
        self.p.read_url = 'postgresql://u2:pass@localhost/ds'
        assert not self.p._same_read_and_write_url()

    def test_same_ckan_and_datastore_db(self):
        self.p.read_url = 'postgresql://u2:pass@localhost/ckan'
        self.p.ckan_url = 'postgresql://u:pass@localhost/ckan'
        assert self.p._same_ckan_and_datastore_db()

        self.p.read_url = 'postgresql://u:pass@localhost/dt'
        self.p.ckan_url = 'postgresql://u:pass@localhost/ckan'
        assert not self.p._same_ckan_and_datastore_db()

    def test_setup_plugin_for_check_urls_and_permissions_tests_should_leave_the_plugin_in_a_valid_state(self):
        self.setUp_plugin_for_check_urls_and_permissions_tests()
        self.p._check_urls_and_permissions()  # Should be OK

    def test_check_urls_and_permissions_requires_different_ckan_and_datastore_dbs(self):
        self.setUp_plugin_for_check_urls_and_permissions_tests()

        self.p._same_ckan_and_datastore_db = lambda: False
        self.p._check_urls_and_permissions()  # Should be OK

        self.p._same_ckan_and_datastore_db = lambda: True
        nose.tools.assert_raises(InvalidUrlsOrPermissionsException, self.p._check_urls_and_permissions)

    def test_check_urls_and_permissions_requires_different_read_and_write_urls_when_not_in_legacy_mode(self):
        self.setUp_plugin_for_check_urls_and_permissions_tests()
        self.p.legacy_mode = False

        self.p._same_read_and_write_url = lambda: False
        self.p._check_urls_and_permissions()  # Should be OK

        self.p._same_read_and_write_url = lambda: True
        nose.tools.assert_raises(InvalidUrlsOrPermissionsException, self.p._check_urls_and_permissions)

    def test_check_urls_and_permissions_doesnt_require_different_read_and_write_urls_when_in_legacy_mode(self):
        self.setUp_plugin_for_check_urls_and_permissions_tests()
        self.p.legacy_mode = True

        self.p._same_read_and_write_url = lambda: False
        self.p._check_urls_and_permissions()  # Should be OK

        self.p._same_read_and_write_url = lambda: True
        self.p._check_urls_and_permissions()  # Should be OK

    def test_check_urls_and_permissions_requires_read_connection_with_correct_privileges_when_not_in_legacy_mode(self):
        self.setUp_plugin_for_check_urls_and_permissions_tests()
        self.p.legacy_mode = False

        self.p._read_connection_has_correct_privileges = lambda: True
        self.p._check_urls_and_permissions()  # Should be OK

        self.p._read_connection_has_correct_privileges = lambda: False
        nose.tools.assert_raises(InvalidUrlsOrPermissionsException, self.p._check_urls_and_permissions)

    def test_check_urls_and_permissions_doesnt_care_about_read_connection_privileges_when_in_legacy_mode(self):
        self.setUp_plugin_for_check_urls_and_permissions_tests()
        self.p.legacy_mode = True

        self.p._read_connection_has_correct_privileges = lambda: True
        self.p._check_urls_and_permissions()  # Should be OK

        self.p._read_connection_has_correct_privileges = lambda: False
        self.p._check_urls_and_permissions()  # Should be OK

    def setUp_plugin_for_check_urls_and_permissions_tests(self):
        def _raise_invalid_urls_or_permissions_exception(message):
            raise InvalidUrlsOrPermissionsException(message)

        self.p._same_ckan_and_datastore_db = lambda: False
        self.p.legacy_mode = True
        self.p._same_read_and_write_url = lambda: False
        self.p._read_connection_has_correct_privileges = lambda: True
        self.p._log_or_raise = _raise_invalid_urls_or_permissions_exception


class InvalidUrlsOrPermissionsException(Exception):
    pass



# encoding: utf-8

from routes import url_for

import ckan.tests.helpers as helpers
import ckan.tests.factories as factories


class TestFeedNew(helpers.FunctionalTestBase):

    def test_atom_feed_page_zero_gives_error(self):
        group = factories.Group()
        offset = url_for(controller='feed', action='group',
                         id=group['name']) + '?page=0'
        app = self._get_test_app()
        res = app.get(offset, status=400)
        assert '"page" parameter must be a positive integer' in res, res

    def test_atom_feed_page_negative_gives_error(self):
        group = factories.Group()
        offset = url_for(controller='feed', action='group',
                         id=group['name']) + '?page=-2'
        app = self._get_test_app()
        res = app.get(offset, status=400)
        assert '"page" parameter must be a positive integer' in res, res

    def test_atom_feed_page_not_int_gives_error(self):
        group = factories.Group()
        offset = url_for(controller='feed', action='group',
                         id=group['name']) + '?page=abc'
        app = self._get_test_app()
        res = app.get(offset, status=400)
        assert '"page" parameter must be a positive integer' in res, res



# encoding: utf-8

'''
The aim of these tests is to check and improve the coding standards in ckan.
Common issues are tested for here and tests fail if they are discovered in
files that are either new or were previously good. Bad files are
blacklisted to prevent them throwing errors in many cases because of the
number of affected files e.g. PEP8.  However if files start to pass a test
will fail and the file should be removed from the blacklist so that it will
then be kept clean in future.

The idea is to slowly improve the code quality in ckan without having files
deteriourating when they do reach the required standard.

Please do not add new files to the list as any new files should meet the
current coding standards.  Please add comments by files that fail if there
are legitimate reasons for the failure.
'''

import sys
import os
import re
import cStringIO
import inspect
import itertools

import pep8

file_path = os.path.dirname(__file__)
base_path = os.path.abspath(os.path.join(file_path, '..', '..', '..'))


def process_directory(directory, ext='.py'):
    base_len = len(base_path) + 1
    for (dirpath, dirnames, filenames) in os.walk(directory):
        # ignore hidden files and dir
        filenames = [f for f in filenames if not f[0] == '.']
        dirnames[:] = [d for d in dirnames if not d[0] == '.']
        for name in filenames:
            if name.endswith(ext):
                path = os.path.join(dirpath, name)
                filename = path[base_len:]
                yield path, filename


def output_errors(filename, errors):
    out = ['']
    out.append('-' * len(filename))
    out.append(filename)
    out.append('-' * len(filename))
    for error in errors:
        out.append(error)
    return '\n'.join(out)


def show_fails(msg, errors):
    if errors:
        msg = ['\n%s' % msg]
        for error in errors:
            msg.append(errors[error])
        msg.append('\n\nFailing Files:\n==============')
        msg += sorted(errors)
        raise Exception('\n'.join(msg))


def show_passing(msg, errors):
    if errors:
        raise Exception('\n%s\n\n' % msg + '\n'.join(sorted(errors)))


def cs_filter(f, filter_, ignore_comment_lines=True):
    ''' filter the file removing comments if requested.
    looks for comments like
    # CS: <filter_> ignore
    # CS: <filter_> ignore x line
    and removes the requested number of lines.  Lines are removed by
    blanking so the line numbers reported will be correct.  This allows us
    to check files that have known violations of the test rules. '''

    # this RegEx is of poor quality but works
    exp = r'^\s*#\s+CS:.*%s.*ignore\D*((\d+)\s+line)*'
    re_ignore = re.compile(exp % filter_)
    ignore = 0
    out = []
    count = 1
    for line in f:
        # ignore the line if we have been told too
        if ignore > 0:
            line = ''
            ignore -= 1
        matches = re_ignore.search(line)
        if matches:
            ignore = int(matches.group(2) or 1)
        # ignore comments out lines
        if ignore_comment_lines and line.lstrip().startswith('#'):
            line = ''
        out.append(line)
        count += 1
    return out


class TestBadSpellings(object):

    BAD_SPELLING_BLACKLIST_FILES = []

    # these are the bad spellings with the correct spelling
    # use LOWER case
    BAD_SPELLINGS = {
        # CS: bad_spelling ignore 2 lines
        'licence': 'license',
        'organisation': 'organization',
    }

    fails = {}
    passes = []
    done = False

    @classmethod
    def setup(cls):
        if not cls.done:
            cls.process()
        cls.done = True

    @classmethod
    def process(cls):
        blacklist = cls.BAD_SPELLING_BLACKLIST_FILES
        re_bad_spelling = re.compile(
            r'(%s)' % '|'.join([x for x in cls.BAD_SPELLINGS]),
            flags=re.IGNORECASE
        )
        files = itertools.chain.from_iterable([
            process_directory(base_path),
            process_directory(base_path, ext='.rst')])
        for path, filename in files:
            f = open(path, 'r')
            count = 1
            errors = []
            for line in cs_filter(f, 'bad_spelling'):
                matches = re_bad_spelling.findall(line)
                if matches:
                    bad_words = []
                    for m in matches:
                        if m not in bad_words:
                            bad_words.append('%s use %s' %
                                             (m, cls.BAD_SPELLINGS[m.lower()]))
                    bad = ', '.join(bad_words)
                    errors.append('ln:%s \t%s\n<%s>' % (count, line[:-1], bad))
                count += 1
            if errors and filename not in blacklist:
                cls.fails[filename] = output_errors(filename, errors)
            elif not errors and filename in blacklist:
                cls.passes.append(filename)

    def test_good(self):
        msg = 'The following files passed bad spellings rules'
        msg += '\nThey need removing from the test blacklist'
        show_passing(msg, self.passes)

    def test_bad(self):
        msg = 'The following files have bad spellings that need fixing'
        show_fails(msg, self.fails)


class TestNastyString(object):
    # CS: nasty_string ignore
    ''' Look for a common coding problem in ckan '..%s..' % str(x) '''

    # Nasty str() issues
    #
    # There are places in ckan where code is like `'...%s..' % str(..)`
    # these cause problems when unicode is present but can remain dormant
    # for a long time before the issue is apparent so try to remove these.
    # The value is converted to a string anyway so the str() is unneeded in
    # any place.

    NASTY_STR_BLACKLIST_FILES = []

    fails = {}
    passes = []
    done = False

    @classmethod
    def setup(cls):
        if not cls.done:
            cls.process()
        cls.done = True

    @classmethod
    def process(cls):
        blacklist = cls.NASTY_STR_BLACKLIST_FILES
        re_nasty_str = re.compile(
            r'''("[^"]*\%s[^"]*"|'[^']*\%s[^']*').*%.*str\('''
        )
        for path, filename in process_directory(base_path):
            f = open(path, 'r')
            count = 1
            errors = []
            for line in cs_filter(f, 'nasty_string'):
                if re_nasty_str.search(line):
                    errors.append('ln:%s \t%s' % (count, line[:-1]))
                count += 1
            if errors and filename not in blacklist:
                cls.fails[filename] = output_errors(filename, errors)
            elif not errors and filename in blacklist:
                cls.passes.append(filename)

    def test_good(self):
        msg = 'The following files passed nasty str() rules'
        msg += '\nThey need removing from the test blacklist'
        show_passing(msg, self.passes)

    def test_bad(self):
        # CS: nasty_string ignore next 2 lines
        msg = ('The following files have nasty str() issues that need'
               ' resolving\nCode is like `\'...%s..\' % str(..)`'
               'and should just be `\'...%s..\' % ..`')
        show_fails(msg, self.fails)


class TestImportStar(object):
    ''' Find files using from xxx import * '''

    # Import * file exceptions
    #
    # The following files contain one or more `from ... import *` lines
    # which should not be used in ckan where possible.  If the files get
    # fixed they should be removed from this list.
    #
    # import * is bad for many reasons and should be avoided.

    IMPORT_STAR_BLACKLIST_FILES = [
        'ckan/migration/versions/001_add_existing_tables.py',
        'ckan/migration/versions/002_add_author_and_maintainer.py',
        'ckan/migration/versions/003_add_user_object.py',
        'ckan/migration/versions/004_add_group_object.py',
        'ckan/migration/versions/005_add_authorization_tables.py',
        'ckan/migration/versions/006_add_ratings.py',
        'ckan/migration/versions/007_add_system_roles.py',
        'ckan/migration/versions/008_update_vdm_ids.py',
        'ckan/migration/versions/009_add_creation_timestamps.py',
        'ckan/migration/versions/010_add_user_about.py',
        'ckan/migration/versions/011_add_package_search_vector.py',
        'ckan/migration/versions/012_add_resources.py',
        'ckan/migration/versions/013_add_hash.py',
        'ckan/migration/versions/014_hash_2.py',
        'ckan/migration/versions/015_remove_state_object.py',
        'ckan/migration/versions/016_uuids_everywhere.py',
        'ckan/migration/versions/017_add_pkg_relationships.py',
        'ckan/migration/versions/018_adjust_licenses.py',
        'ckan/migration/versions/019_pkg_relationships_state.py',
        'ckan/migration/versions/020_add_changeset.py',
        'ckan/migration/versions/022_add_group_extras.py',
        'ckan/migration/versions/023_add_harvesting.py',
        'ckan/migration/versions/024_add_harvested_document.py',
        'ckan/migration/versions/025_add_authorization_groups.py',
        'ckan/migration/versions/026_authorization_group_user_pk.py',
        'ckan/migration/versions/027_adjust_harvester.py',
        'ckan/migration/versions/028_drop_harvest_source_status.py',
        'ckan/migration/versions/029_version_groups.py',
        'ckan/migration/versions/030_additional_user_attributes.py',
        'ckan/migration/versions/031_move_openid_to_new_field.py',
        'ckan/migration/versions/032_add_extra_info_field_to_resources.py',
        'ckan/migration/versions/033_auth_group_user_id_add_conditional.py',
        'ckan/migration/versions/034_resource_group_table.py',
        'ckan/migration/versions/035_harvesting_doc_versioning.py',
        'ckan/migration/versions/036_lockdown_roles.py',
        'ckan/migration/versions/037_role_anon_editor.py',
        'ckan/migration/versions/038_delete_migration_tables.py',
        'ckan/migration/versions/039_add_expired_id_and_dates.py',
        'ckan/migration/versions/040_reset_key_on_user.py',
        'ckan/migration/versions/041_resource_new_fields.py',
        'ckan/migration/versions/042_user_revision_indexes.py',
        'ckan/migration/versions/043_drop_postgres_search.py',
        'ckan/migration/versions/044_add_task_status.py',
        'ckan/migration/versions/045_user_name_unique.py',
        'ckan/migration/versions/046_drop_changesets.py',
        'ckan/migration/versions/047_rename_package_group_member.py',
        'ckan/migration/versions/048_add_activity_streams_tables.py',
        'ckan/migration/versions/049_add_group_approval_status.py',
        'ckan/migration/versions/050_term_translation_table.py',
        'ckan/migration/versions/051_add_tag_vocabulary.py',
        'ckan/migration/versions/052_update_member_capacities.py',
        'ckan/migration/versions/053_add_group_logo.py',
        'ckan/migration/versions/056_add_related_table.py',
        'ckan/migration/versions/057_tracking.py',
        'ckan/migration/versions/058_add_follower_tables.py',
        'ckan/migration/versions/059_add_related_count_and_flag.py',
        'ckan/migration/versions/060_add_system_info_table.py',
        'ckan/migration/versions/061_add_follower__group_table.py',
        'ckan/migration/versions/062_add_dashboard_table.py',
        'ckan/migration/versions/063_org_changes.py',
        'ckan/migration/versions/064_add_email_last_sent_column.py',
        'ckan/migration/versions/065_add_email_notifications_preference.py',
        'ckan/plugins/__init__.py',
        'ckan/tests/legacy/functional/api/base.py',
        'ckan/tests/legacy/functional/api/test_api.py',
        'ckan/tests/legacy/functional/api/test_misc.py',
        'ckan/tests/legacy/functional/api/test_package_search.py',
        'ckan/tests/legacy/functional/api/test_resource_search.py',
        'ckan/tests/legacy/functional/api/test_revision_search.py',
        'ckan/tests/legacy/functional/test_group.py',
        'ckan/tests/legacy/functional/test_home.py',
        'ckan/tests/legacy/functional/test_package.py',
        'ckan/tests/legacy/functional/test_package_relationships.py',
        'ckan/tests/legacy/functional/test_tag.py',
        'ckan/tests/legacy/lib/test_helpers.py',
        'ckan/tests/legacy/lib/test_resource_search.py',
        'ckan/tests/legacy/lib/test_tag_search.py',
        'ckan/tests/legacy/misc/test_sync.py',
        'ckan/tests/legacy/models/test_extras.py',
        'ckan/tests/legacy/models/test_misc.py',
        'ckan/tests/legacy/models/test_package.py',
        'ckan/tests/legacy/models/test_package_relationships.py',
        'ckan/tests/legacy/models/test_purge_revision.py',
        'ckan/tests/legacy/models/test_resource.py',
        'ckan/tests/legacy/models/test_revision.py',
        'ckan/tests/legacy/models/test_user.py',
        'ckan/tests/legacy/pylons_controller.py',
        'fabfile.py',
    ]
    fails = {}
    passes = []
    done = False

    @classmethod
    def setup(cls):
        if not cls.done:
            cls.process()
        cls.done = True

    @classmethod
    def process(cls):
        blacklist = cls.IMPORT_STAR_BLACKLIST_FILES
        re_import_star = re.compile(r'^\s*from\s+.*\simport\s+\*')
        for path, filename in process_directory(base_path):
            f = open(path, 'r')
            count = 1
            errors = []
            for line in f:
                if re_import_star.search(line):
                    errors.append('%s ln:%s import *\n\t%s'
                                  % (filename, count, line))
                count += 1
            if errors and filename not in blacklist:
                cls.fails[filename] = output_errors(filename, errors)
            elif not errors and filename in blacklist:
                cls.passes.append(filename)

    def test_import_good(self):
        msg = 'The following files passed import * rules'
        msg += '\nThey need removing from the test blacklist'
        show_passing(msg, self.passes)

    def test_import_bad(self):
        msg = ('The following files have import * issues that need resolving\n'
               '`from ... import *` lines which should not be used in ckan'
               ' where possible.')
        show_fails(msg, self.fails)


class TestPep8(object):
    ''' Check that .py files are pep8 compliant '''

    # PEP8 File exceptions
    #
    # The following files have known PEP8 errors.  When the files get to a
    # point of not having any such errors they should be removed from this
    # list to prevent new errors being added to the file.

    PEP8_BLACKLIST_FILES = [
        'bin/running_stats.py',
        'ckan/__init__.py',
        'ckan/ckan_nose_plugin.py',
        'ckan/config/middleware.py',
        'ckan/config/routing.py',
        'ckan/config/sp_config.py',
        'ckan/controllers/admin.py',
        'ckan/controllers/revision.py',
        'ckan/include/rcssmin.py',
        'ckan/include/rjsmin.py',
        'ckan/lib/activity_streams.py',
        'ckan/lib/activity_streams_session_extension.py',
        'ckan/lib/alphabet_paginate.py',
        'ckan/lib/app_globals.py',
        'ckan/lib/captcha.py',
        'ckan/lib/cli.py',
        'ckan/lib/create_test_data.py',
        'ckan/lib/dictization/__init__.py',
        'ckan/lib/dictization/model_dictize.py',
        'ckan/lib/dictization/model_save.py',
        'ckan/lib/email_notifications.py',
        'ckan/lib/extract.py',
        'ckan/lib/fanstatic_extensions.py',
        'ckan/lib/fanstatic_resources.py',
        'ckan/lib/formatters.py',
        'ckan/lib/hash.py',
        'ckan/lib/help/flash_messages.py',
        'ckan/lib/jinja_extensions.py',
        'ckan/lib/jsonp.py',
        'ckan/lib/maintain.py',
        'ckan/lib/navl/validators.py',
        'ckan/lib/package_saver.py',
        'ckan/lib/plugins.py',
        'ckan/lib/render.py',
        'ckan/lib/search/__init__.py',
        'ckan/lib/search/index.py',
        'ckan/lib/search/query.py',
        'ckan/lib/search/sql.py',
        'ckan/logic/action/__init__.py',
        'ckan/logic/action/delete.py',
        'ckan/logic/action/get.py',
        'ckan/logic/action/update.py',
        'ckan/logic/auth/create.py',
        'ckan/logic/auth/delete.py',
        'ckan/logic/auth/get.py',
        'ckan/logic/auth/update.py',
        'ckan/logic/converters.py',
        'ckan/logic/validators.py',
        'ckan/migration/versions/001_add_existing_tables.py',
        'ckan/migration/versions/002_add_author_and_maintainer.py',
        'ckan/migration/versions/003_add_user_object.py',
        'ckan/migration/versions/004_add_group_object.py',
        'ckan/migration/versions/005_add_authorization_tables.py',
        'ckan/migration/versions/006_add_ratings.py',
        'ckan/migration/versions/007_add_system_roles.py',
        'ckan/migration/versions/008_update_vdm_ids.py',
        'ckan/migration/versions/009_add_creation_timestamps.py',
        'ckan/migration/versions/010_add_user_about.py',
        'ckan/migration/versions/011_add_package_search_vector.py',
        'ckan/migration/versions/012_add_resources.py',
        'ckan/migration/versions/013_add_hash.py',
        'ckan/migration/versions/014_hash_2.py',
        'ckan/migration/versions/015_remove_state_object.py',
        'ckan/migration/versions/016_uuids_everywhere.py',
        'ckan/migration/versions/017_add_pkg_relationships.py',
        'ckan/migration/versions/018_adjust_licenses.py',
        'ckan/migration/versions/019_pkg_relationships_state.py',
        'ckan/migration/versions/020_add_changeset.py',
        'ckan/migration/versions/022_add_group_extras.py',
        'ckan/migration/versions/023_add_harvesting.py',
        'ckan/migration/versions/024_add_harvested_document.py',
        'ckan/migration/versions/025_add_authorization_groups.py',
        'ckan/migration/versions/026_authorization_group_user_pk.py',
        'ckan/migration/versions/027_adjust_harvester.py',
        'ckan/migration/versions/028_drop_harvest_source_status.py',
        'ckan/migration/versions/029_version_groups.py',
        'ckan/migration/versions/030_additional_user_attributes.py',
        'ckan/migration/versions/031_move_openid_to_new_field.py',
        'ckan/migration/versions/032_add_extra_info_field_to_resources.py',
        'ckan/migration/versions/033_auth_group_user_id_add_conditional.py',
        'ckan/migration/versions/034_resource_group_table.py',
        'ckan/migration/versions/035_harvesting_doc_versioning.py',
        'ckan/migration/versions/036_lockdown_roles.py',
        'ckan/migration/versions/037_role_anon_editor.py',
        'ckan/migration/versions/038_delete_migration_tables.py',
        'ckan/migration/versions/039_add_expired_id_and_dates.py',
        'ckan/migration/versions/040_reset_key_on_user.py',
        'ckan/migration/versions/041_resource_new_fields.py',
        'ckan/migration/versions/042_user_revision_indexes.py',
        'ckan/migration/versions/043_drop_postgres_search.py',
        'ckan/migration/versions/044_add_task_status.py',
        'ckan/migration/versions/045_user_name_unique.py',
        'ckan/migration/versions/046_drop_changesets.py',
        'ckan/migration/versions/047_rename_package_group_member.py',
        'ckan/migration/versions/048_add_activity_streams_tables.py',
        'ckan/migration/versions/049_add_group_approval_status.py',
        'ckan/migration/versions/050_term_translation_table.py',
        'ckan/migration/versions/051_add_tag_vocabulary.py',
        'ckan/migration/versions/052_update_member_capacities.py',
        'ckan/migration/versions/053_add_group_logo.py',
        'ckan/migration/versions/054_add_resource_created_date.py',
        'ckan/migration/versions/055_update_user_and_activity_detail.py',
        'ckan/migration/versions/056_add_related_table.py',
        'ckan/migration/versions/057_tracking.py',
        'ckan/migration/versions/058_add_follower_tables.py',
        'ckan/migration/versions/059_add_related_count_and_flag.py',
        'ckan/migration/versions/060_add_system_info_table.py',
        'ckan/migration/versions/061_add_follower__group_table.py',
        'ckan/migration/versions/062_add_dashboard_table.py',
        'ckan/migration/versions/063_org_changes.py',
        'ckan/migration/versions/064_add_email_last_sent_column.py',
        'ckan/migration/versions/065_add_email_notifications_preference.py',
        'ckan/migration/versions/067_turn_extras_to_strings.py',
        'ckan/misc.py',
        'ckan/model/__init__.py',
        'ckan/model/activity.py',
        'ckan/model/authz.py',
        'ckan/model/dashboard.py',
        'ckan/model/domain_object.py',
        'ckan/model/follower.py',
        'ckan/model/group.py',
        'ckan/model/group_extra.py',
        'ckan/model/license.py',
        'ckan/model/meta.py',
        'ckan/model/misc.py',
        'ckan/model/modification.py',
        'ckan/model/package.py',
        'ckan/model/package_extra.py',
        'ckan/model/package_relationship.py',
        'ckan/model/rating.py',
        'ckan/model/resource.py',
        'ckan/model/system_info.py',
        'ckan/model/tag.py',
        'ckan/model/task_status.py',
        'ckan/model/term_translation.py',
        'ckan/model/test_user.py',
        'ckan/model/tracking.py',
        'ckan/model/types.py',
        'ckan/model/user.py',
        'ckan/model/vocabulary.py',
        'ckan/authz.py',
        'ckan/pastertemplates/__init__.py',
        'ckan/plugins/interfaces.py',
        'ckan/poo.py',
        'ckan/rating.py',
        'ckan/templates_legacy/home/__init__.py',
        'ckan/tests/legacy/__init__.py',
        'ckan/tests/legacy/ckantestplugin/ckantestplugin/__init__.py',
        'ckan/tests/legacy/ckantestplugin/setup.py',
        'ckan/tests/legacy/ckantestplugins.py',
        'ckan/tests/legacy/functional/api/base.py',
        'ckan/tests/legacy/functional/api/model/test_group.py',
        'ckan/tests/legacy/functional/api/model/test_licenses.py',
        'ckan/tests/legacy/functional/api/model/test_package.py',
        'ckan/tests/legacy/functional/api/model/test_ratings.py',
        'ckan/tests/legacy/functional/api/model/test_relationships.py',
        'ckan/tests/legacy/functional/api/model/test_revisions.py',
        'ckan/tests/legacy/functional/api/model/test_tag.py',
        'ckan/tests/legacy/functional/api/model/test_vocabulary.py',
        'ckan/tests/legacy/functional/api/test_activity.py',
        'ckan/tests/legacy/functional/api/test_api.py',
        'ckan/tests/legacy/functional/api/test_dashboard.py',
        'ckan/tests/legacy/functional/api/test_email_notifications.py',
        'ckan/tests/legacy/functional/api/test_follow.py',
        'ckan/tests/legacy/functional/api/test_misc.py',
        'ckan/tests/legacy/functional/api/test_package_search.py',
        'ckan/tests/legacy/functional/api/test_resource.py',
        'ckan/tests/legacy/functional/api/test_resource_search.py',
        'ckan/tests/legacy/functional/api/test_revision_search.py',
        'ckan/tests/legacy/functional/api/test_user.py',
        'ckan/tests/legacy/functional/api/test_util.py',
        'ckan/tests/legacy/functional/base.py',
        'ckan/tests/legacy/functional/test_activity.py',
        'ckan/tests/legacy/functional/test_admin.py',
        'ckan/tests/legacy/functional/test_cors.py',
        'ckan/tests/legacy/functional/test_error.py',
        'ckan/tests/legacy/functional/test_home.py',
        'ckan/tests/legacy/functional/test_package.py',
        'ckan/tests/legacy/functional/test_package_relationships.py',
        'ckan/tests/legacy/functional/test_pagination.py',
        'ckan/tests/legacy/functional/test_preview_interface.py',
        'ckan/tests/legacy/functional/test_revision.py',
        'ckan/tests/legacy/functional/test_search.py',
        'ckan/tests/legacy/functional/test_tag.py',
        'ckan/tests/legacy/functional/test_tag_vocab.py',
        'ckan/tests/legacy/functional/test_upload.py',
        'ckan/tests/legacy/functional/test_user.py',
        'ckan/tests/legacy/html_check.py',
        'ckan/tests/legacy/lib/__init__.py',
        'ckan/tests/legacy/lib/test_accept.py',
        'ckan/tests/legacy/lib/test_cli.py',
        'ckan/tests/legacy/lib/test_dictization.py',
        'ckan/tests/legacy/lib/test_email_notifications.py',
        'ckan/tests/legacy/lib/test_hash.py',
        'ckan/tests/legacy/lib/test_helpers.py',
        'ckan/tests/legacy/lib/test_mailer.py',
        'ckan/tests/legacy/lib/test_munge.py',
        'ckan/tests/legacy/lib/test_navl.py',
        'ckan/tests/legacy/lib/test_resource_search.py',
        'ckan/tests/legacy/lib/test_simple_search.py',
        'ckan/tests/legacy/lib/test_solr_package_search.py',
        'ckan/tests/legacy/lib/test_solr_package_search_synchronous_update.py',
        'ckan/tests/legacy/lib/test_solr_schema_version.py',
        'ckan/tests/legacy/lib/test_solr_search_index.py',
        'ckan/tests/legacy/lib/test_tag_search.py',
        'ckan/tests/legacy/logic/test_action.py',
        'ckan/tests/legacy/logic/test_auth.py',
        'ckan/tests/legacy/logic/test_tag.py',
        'ckan/tests/legacy/logic/test_validators.py',
        'ckan/tests/legacy/misc/test_format_text.py',
        'ckan/tests/legacy/misc/test_mock_mail_server.py',
        'ckan/tests/legacy/misc/test_sync.py',
        'ckan/tests/legacy/mock_plugin.py',
        'ckan/tests/legacy/models/test_extras.py',
        'ckan/tests/legacy/models/test_group.py',
        'ckan/tests/legacy/models/test_license.py',
        'ckan/tests/legacy/models/test_misc.py',
        'ckan/tests/legacy/models/test_package.py',
        'ckan/tests/legacy/models/test_package_relationships.py',
        'ckan/tests/legacy/models/test_purge_revision.py',
        'ckan/tests/legacy/models/test_resource.py',
        'ckan/tests/legacy/models/test_revision.py',
        'ckan/tests/legacy/models/test_user.py',
        'ckan/tests/legacy/monkey.py',
        'ckan/tests/legacy/pylons_controller.py',
        'ckan/tests/legacy/schema/test_schema.py',
        'ckan/tests/legacy/test_plugins.py',
        'ckan/tests/legacy/test_versions.py',
        'ckan/websetup.py',
        'ckanext/datastore/bin/datastore_setup.py',
        'ckanext/datastore/logic/action.py',
        'ckanext/datastore/plugin.py',
        'ckanext/datastore/tests/test_create.py',
        'ckanext/datastore/tests/test_search.py',
        'ckanext/datastore/tests/test_upsert.py',
        'ckanext/example_idatasetform/plugin.py',
        'ckanext/example_itemplatehelpers/plugin.py',
        'ckanext/multilingual/plugin.py',
        'ckanext/resourceproxy/plugin.py',
        'ckanext/stats/controller.py',
        'ckanext/stats/plugin.py',
        'ckanext/stats/stats.py',
        'ckanext/stats/tests/__init__.py',
        'ckanext/stats/tests/test_stats_lib.py',
        'ckanext/stats/tests/test_stats_plugin.py',
        'ckanext/test_tag_vocab_plugin.py',
        'ckanext/tests/plugin.py',
        'doc/conf.py',
        'fabfile.py',
        'profile_tests.py',
        'setup.py',
    ]
    fails = {}
    passes = []
    done = False

    @classmethod
    def setup(cls):
        if not cls.done:
            cls.process()
        cls.done = True

    @classmethod
    def process(cls):
        blacklist = cls.PEP8_BLACKLIST_FILES
        for path, filename in process_directory(base_path):
            errors = cls.find_pep8_errors(filename=path)
            if errors and filename not in blacklist:
                cls.fails[filename] = output_errors(filename, errors)
            elif not errors and filename in blacklist:
                cls.passes.append(filename)

    def test_pep8_fails(self):
        msg = 'The following files have pep8 issues that need resolving'
        msg += '\nThey need removing from the test blacklist'
        show_fails(msg, self.fails)

    def test_pep8_pass(self):
        msg = 'The following files passed pep8 but are blacklisted'
        show_passing(msg, self.passes)

    @classmethod
    def find_pep8_errors(cls, filename=None, lines=None):
        try:
            sys.stdout = cStringIO.StringIO()
            config = {}

            # Ignore long lines on test files, as the test names can get long
            # when following our test naming standards.
            if cls._is_test(filename):
                config['ignore'] = ['E501']

            checker = pep8.Checker(filename=filename, lines=lines,
                                   **config)
            checker.check_all()
            output = sys.stdout.getvalue()
        finally:
            sys.stdout = sys.__stdout__

        errors = []
        for line in output.split('\n'):
            parts = line.split(' ', 2)
            if len(parts) == 3:
                location, error, desc = parts
                line_no = location.split(':')[1]
                errors.append('%s ln:%s %s' % (error, line_no, desc))
        return errors

    @classmethod
    def _is_test(cls, filename):
        return bool(re.search('(^|\W)test_.*\.py$', filename, re.IGNORECASE))


class TestActionAuth(object):
    ''' These tests check the logic auth/action functions are compliant. The
    main tests are that each action has a corresponding auth function and
    that each auth function has an action.  We check the function only
    accepts (context, data_dict) as parameters. '''

    ACTION_FN_SIGNATURES_BLACKLIST = [
        'create: activity_create',
    ]

    ACTION_NO_AUTH_BLACKLIST = [
        'create: follow_dataset',
        'create: follow_group',
        'create: follow_user',
        'create: package_relationship_create_rest',
        'delete: package_relationship_delete_rest',
        'delete: unfollow_dataset',
        'delete: unfollow_group',
        'delete: unfollow_user',
        'get: activity_detail_list',
        'get: am_following_dataset',
        'get: am_following_group',
        'get: am_following_user',
        'get: dashboard_activity_list_html',
        'get: dataset_followee_count',
        'get: dataset_follower_count',
        'get: followee_count',
        'get: group_activity_list',
        'get: group_activity_list_html',
        'get: group_followee_count',
        'get: group_follower_count',
        'get: group_package_show',
        'get: member_list',
        'get: organization_activity_list',
        'get: organization_activity_list_html',
        'get: organization_follower_count',
        'get: package_activity_list',
        'get: package_activity_list_html',
        'get: recently_changed_packages_activity_list',
        'get: recently_changed_packages_activity_list_html',
        'get: resource_search',
        'get: roles_show',
        'get: status_show',
        'get: tag_search',
        'get: term_translation_show',
        'get: user_activity_list',
        'get: user_activity_list_html',
        'get: user_followee_count',
        'get: user_follower_count',
        'update: package_relationship_update_rest',
        'update: task_status_update_many',
        'update: term_translation_update_many',
    ]

    AUTH_NO_ACTION_BLACKLIST = [
        'create: file_upload',
        'delete: revision_delete',
        'delete: revision_undelete',
        'get: group_autocomplete',
        'get: group_list_available',
        'get: sysadmin',
        'get: request_reset',
        'get: user_reset',
        'update: group_change_state',
        'update: group_edit_permissions',
        'update: package_change_state',
        'update: revision_change_state',
    ]

    ACTION_NO_DOC_STR_BLACKLIST = [
        'create: group_create_rest',
        'create: package_create_rest',
        'create: package_relationship_create_rest',
        'delete: package_relationship_delete_rest',
        'get: get_site_user',
        'get: group_show_rest',
        'get: package_show_rest',
        'get: tag_show_rest',
        'update: group_update_rest',
        'update: package_relationship_update_rest',
        'update: package_update_rest',
    ]

    done = False

    @classmethod
    def setup(cls):
        if not cls.done:
            cls.process()
        cls.done = True

    @classmethod
    def process(cls):
        def get_functions(module_root):
            fns = {}
            for auth_module_name in ['get', 'create', 'update', 'delete', 'patch']:
                module_path = '%s.%s' % (module_root, auth_module_name,)
                try:
                    module = __import__(module_path)
                except ImportError:
                    print ('No auth module for action "%s"' % auth_module_name)

                for part in module_path.split('.')[1:]:
                    module = getattr(module, part)

                for key, v in module.__dict__.items():
                    if not hasattr(v, '__call__'):
                        continue
                    if v.__module__ != module_path:
                        continue
                    if not key.startswith('_'):
                        name = '%s: %s' % (auth_module_name, key)
                        fns[name] = v
            return fns
        cls.actions = get_functions('logic.action')
        cls.auths = get_functions('logic.auth')

    def test_actions_have_auth_fn(self):
        actions_no_auth = set(self.actions.keys()) - set(self.auths.keys())
        actions_no_auth -= set(self.ACTION_NO_AUTH_BLACKLIST)
        assert not actions_no_auth, 'These actions have no auth function\n%s' \
            % '\n'.join(sorted(list(actions_no_auth)))

    def test_actions_have_auth_fn_blacklist(self):
        actions_no_auth = set(self.actions.keys()) & set(self.auths.keys())
        actions_no_auth &= set(self.ACTION_NO_AUTH_BLACKLIST)
        assert not actions_no_auth, 'These actions blacklisted but ' + \
            'shouldn\'t be \n%s' % '\n'.join(sorted(list(actions_no_auth)))

    def test_auths_have_action_fn(self):
        auths_no_action = set(self.auths.keys()) - set(self.actions.keys())
        auths_no_action -= set(self.AUTH_NO_ACTION_BLACKLIST)
        assert not auths_no_action, 'These auth functions have no action\n%s' \
            % '\n'.join(sorted(list(auths_no_action)))

    def test_auths_have_action_fn_blacklist(self):
        auths_no_action = set(self.auths.keys()) & set(self.actions.keys())
        auths_no_action &= set(self.AUTH_NO_ACTION_BLACKLIST)
        assert not auths_no_action, 'These auths functions blacklisted but' + \
            ' shouldn\'t be \n%s' % '\n'.join(sorted(list(auths_no_action)))

    def test_fn_signatures(self):
        errors = []
        for name, fn in self.actions.iteritems():
            args_info = inspect.getargspec(fn)
            if (args_info.args != ['context', 'data_dict']
                    or args_info.varargs is not None
                    or args_info.keywords is not None):
                if name not in self.ACTION_FN_SIGNATURES_BLACKLIST:
                    errors.append(name)
        assert not errors, 'These action functions have the wrong function' + \
            ' signature, should be (context, data_dict)\n%s' \
            % '\n'.join(sorted(errors))

    def test_fn_docstrings(self):
        errors = []
        for name, fn in self.actions.iteritems():
            if not getattr(fn, '__doc__', None):
                if name not in self.ACTION_NO_DOC_STR_BLACKLIST:
                    errors.append(name)
        assert not errors, 'These action functions need docstrings\n%s' \
            % '\n'.join(sorted(errors))


class TestBadExceptions(object):
    ''' Look for a common coding problem in ckan Exception(_'...') '''

    # Exceptions should not on the whole be translated as they are for
    # programmers to read in trace backs or log files.  However some like
    # Invalid used in validation functions do get passed back up to the user
    # and so should be translated.

    NASTY_EXCEPTION_BLACKLIST_FILES = [
        'ckan/controllers/api.py',
        'ckan/controllers/user.py',
        'ckan/lib/mailer.py',
        'ckan/logic/action/create.py',
        'ckan/logic/action/delete.py',
        'ckan/logic/action/get.py',
        'ckan/logic/action/update.py',
        'ckan/logic/auth/create.py',
        'ckan/logic/auth/delete.py',
        'ckan/logic/auth/get.py',
        'ckan/authz.py',
        'ckanext/datastore/logic/action.py',
    ]
    fails = {}
    passes = []
    done = False

    @classmethod
    def setup(cls):
        if not cls.done:
            cls.process()
        cls.done = True

    @classmethod
    def process(cls):
        blacklist = cls.NASTY_EXCEPTION_BLACKLIST_FILES
        re_nasty_exception = re.compile(
            r'''raise\W+(?![^I]*Invalid\().*_\('''
        )
        for path, filename in process_directory(base_path):
            f = open(path, 'r')
            count = 1
            errors = []
            for line in f:
                if re_nasty_exception.search(line):
                    errors.append('ln:%s \t%s' % (count, line[:-1]))
                count += 1
            if errors and filename not in blacklist:
                cls.fails[filename] = output_errors(filename, errors)
            elif not errors and filename in blacklist:
                cls.passes.append(filename)

    def test_good(self):
        msg = 'The following files passed nasty exceptions rules'
        msg += '\nThey need removing from the test blacklist'
        show_passing(msg, self.passes)

    def test_bad(self):
        msg = ('The following files have nasty exception issues that need'
               ' resolving\nWe should not be translating exceptions in most'
               ' situations.  We need to when the exception message is passed'
               ' to the front end for example validation')
        show_fails(msg, self.fails)



# encoding: utf-8

'''
Code for generating email notifications for users (e.g. email notifications for
new activities in your dashboard activity stream) and emailing them to the
users.

'''
import datetime
import re

import ckan.model as model
import ckan.logic as logic
import ckan.lib.base as base

from ckan.common import ungettext, config


def string_to_timedelta(s):
    '''Parse a string s and return a standard datetime.timedelta object.

    Handles days, hours, minutes, seconds, and microseconds.

    Accepts strings in these formats:

    2 days
    14 days
    4:35:00 (hours, minutes and seconds)
    4:35:12.087465 (hours, minutes, seconds and microseconds)
    7 days, 3:23:34
    7 days, 3:23:34.087465
    .087465 (microseconds only)

    :raises ckan.logic.ValidationError: if the given string does not match any
        of the recognised formats

    '''
    patterns = []
    days_only_pattern = '(?P<days>\d+)\s+day(s)?'
    patterns.append(days_only_pattern)
    hms_only_pattern = '(?P<hours>\d?\d):(?P<minutes>\d\d):(?P<seconds>\d\d)'
    patterns.append(hms_only_pattern)
    ms_only_pattern = '.(?P<milliseconds>\d\d\d)(?P<microseconds>\d\d\d)'
    patterns.append(ms_only_pattern)
    hms_and_ms_pattern = hms_only_pattern + ms_only_pattern
    patterns.append(hms_and_ms_pattern)
    days_and_hms_pattern = '{0},\s+{1}'.format(days_only_pattern,
            hms_only_pattern)
    patterns.append(days_and_hms_pattern)
    days_and_hms_and_ms_pattern = days_and_hms_pattern + ms_only_pattern
    patterns.append(days_and_hms_and_ms_pattern)

    for pattern in patterns:
        match = re.match('^{0}$'.format(pattern), s)
        if match:
            break

    if not match:
        raise logic.ValidationError('Not a valid time: {0}'.format(s))

    gd = match.groupdict()
    days = int(gd.get('days', '0'))
    hours = int(gd.get('hours', '0'))
    minutes = int(gd.get('minutes', '0'))
    seconds = int(gd.get('seconds', '0'))
    milliseconds = int(gd.get('milliseconds', '0'))
    microseconds = int(gd.get('microseconds', '0'))
    delta = datetime.timedelta(days=days, hours=hours, minutes=minutes,
            seconds=seconds, milliseconds=milliseconds,
            microseconds=microseconds)
    return delta


def _notifications_for_activities(activities, user_dict):
    '''Return one or more email notifications covering the given activities.

    This function handles grouping multiple activities into a single digest
    email.

    :param activities: the activities to consider
    :type activities: list of activity dicts like those returned by
        ckan.logic.action.get.dashboard_activity_list()

    :returns: a list of email notifications
    :rtype: list of dicts each with keys 'subject' and 'body'

    '''
    if not activities:
        return []

    if not user_dict.get('activity_streams_email_notifications'):
        return []

    # We just group all activities into a single "new activity" email that
    # doesn't say anything about _what_ new activities they are.
    # TODO: Here we could generate some smarter content for the emails e.g.
    # say something about the contents of the activities, or single out
    # certain types of activity to be sent in their own individual emails,
    # etc.
    subject = ungettext(
        "{n} new activity from {site_title}",
        "{n} new activities from {site_title}",
        len(activities)).format(
                site_title=config.get('ckan.site_title'),
                n=len(activities))
    body = base.render(
            'activity_streams/activity_stream_email_notifications.text',
            extra_vars={'activities': activities})
    notifications = [{
        'subject': subject,
        'body': body
        }]

    return notifications


def _notifications_from_dashboard_activity_list(user_dict, since):
    '''Return any email notifications from the given user's dashboard activity
    list since `since`.

    '''
    # Get the user's dashboard activity stream.
    context = {'model': model, 'session': model.Session,
            'user': user_dict['id']}
    activity_list = logic.get_action('dashboard_activity_list')(context, {})

    # Filter out the user's own activities., so they don't get an email every
    # time they themselves do something (we are not Trac).
    activity_list = [activity for activity in activity_list
            if activity['user_id'] != user_dict['id']]

    # Filter out the old activities.
    strptime = datetime.datetime.strptime
    fmt = '%Y-%m-%dT%H:%M:%S.%f'
    activity_list = [activity for activity in activity_list
            if strptime(activity['timestamp'], fmt) > since]

    return _notifications_for_activities(activity_list, user_dict)


# A list of functions that provide email notifications for users from different
# sources. Add to this list if you want to implement a new source of email
# notifications.
_notifications_functions = [
    _notifications_from_dashboard_activity_list,
    ]


def get_notifications(user_dict, since):
    '''Return any email notifications for the given user since `since`.

    For example email notifications about activity streams will be returned for
    any activities the occurred since `since`.

    :param user_dict: a dictionary representing the user, should contain 'id'
        and 'name'
    :type user_dict: dictionary

    :param since: datetime after which to return notifications from
    :rtype since: datetime.datetime

    :returns: a list of email notifications
    :rtype: list of dicts with keys 'subject' and 'body'

    '''
    notifications = []
    for function in _notifications_functions:
        notifications.extend(function(user_dict, since))
    return notifications


def send_notification(user, email_dict):
    '''Email `email_dict` to `user`.'''
    import ckan.lib.mailer

    if not user.get('email'):
        # FIXME: Raise an exception.
        return

    try:
        ckan.lib.mailer.mail_recipient(user['display_name'], user['email'],
                email_dict['subject'], email_dict['body'])
    except ckan.lib.mailer.MailerException:
        raise


def get_and_send_notifications_for_user(user):

    # Parse the email_notifications_since config setting, email notifications
    # from longer ago than this time will not be sent.
    email_notifications_since = config.get(
            'ckan.email_notifications_since', '2 days')
    email_notifications_since = string_to_timedelta(
            email_notifications_since)
    email_notifications_since = (datetime.datetime.now()
            - email_notifications_since)

    # FIXME: We are accessing model from lib here but I'm not sure what
    # else to do unless we add a get_email_last_sent() logic function which
    # would only be needed by this lib.
    email_last_sent = model.Dashboard.get(user['id']).email_last_sent
    activity_stream_last_viewed = (
            model.Dashboard.get(user['id']).activity_stream_last_viewed)

    since = max(email_notifications_since, email_last_sent,
            activity_stream_last_viewed)

    notifications = get_notifications(user, since)
    # TODO: Handle failures from send_email_notification.
    for notification in notifications:
        send_notification(user, notification)

    # FIXME: We are accessing model from lib here but I'm not sure what
    # else to do unless we add a update_email_last_sent()
    # logic function which would only be needed by this lib.
    dash = model.Dashboard.get(user['id'])
    dash.email_last_sent = datetime.datetime.now()
    model.repo.commit()


def get_and_send_notifications_for_all_users():
    context = {'model': model, 'session': model.Session, 'ignore_auth': True,
            'keep_email': True}
    users = logic.get_action('user_list')(context, {})
    for user in users:
        get_and_send_notifications_for_user(user)



# encoding: utf-8

from sqlalchemy import *
from migrate import *
import migrate.changeset
import vdm.sqlalchemy



def upgrade(migrate_engine):
    metadata = MetaData()
    metadata.bind = migrate_engine

    package_table = Table('package', metadata, autoload=True)
    package_revision_table = Table('package_revision', metadata, autoload=True)
    package_resource_table = Table('package_resource', metadata,
        Column('id', Integer, primary_key=True),
        Column('package_id', Integer, ForeignKey('package.id')),
        Column('url', UnicodeText, nullable=False),
        Column('format', UnicodeText),
        Column('description', UnicodeText),
        Column('position', Integer),
        Column('state_id', Integer),
        Column('revision_id', UnicodeText, ForeignKey('revision.id'))
        )

    package_resource_revision_table = Table('package_resource_revision', metadata,
        Column('id', Integer, primary_key=True),
        Column('package_id', Integer, ForeignKey('package.id')),
        Column('url', UnicodeText, nullable=False),
        Column('format', UnicodeText),
        Column('description', UnicodeText),
        Column('position', Integer),
        Column('state_id', Integer),
        Column('revision_id', UnicodeText, ForeignKey('revision.id'), primary_key=True),
        Column('continuity_id', Integer, ForeignKey('package_resource.id'))
        )

    
    package_resource_table.create()
    package_resource_revision_table.create()
    
    # Move download_urls across to resources
    # NB: strictly we should check each package_revision to check whether
    # download_url changed (and if only change) and then create
    # package_resource_revision for each such revision (and delete every
    # package_revision where only change is download_url)
    # However, we adopt a cruder approach in which we just create 
    engine = migrate_engine
    select_sql = select([package_table])
    for pkg in engine.execute(select_sql):
        download_url = pkg['download_url']
        if download_url:
            # what about revision_id?
            res_values = {'package_id':pkg.id,
                          'url':download_url,
                          'format':u'',
                          'description':u'',
                          'position':0,
                          'state_id':1,
                          'revision_id': pkg.revision_id,
                          }
            insert_sql = package_resource_table.insert(values=res_values)
            engine.execute(insert_sql)
            # get id of just inserted resource
            getid_sql = select([package_resource_table]).where(
                    package_resource_table.c.package_id==pkg.id)
            resource_id = getid_sql.execute().fetchone().id
            # now we need to update revision table ...
            res_rev_values = dict(res_values)
            res_rev_values['continuity_id'] = resource_id
            res_rev_values['id'] = resource_id
            insert_sql = package_resource_revision_table.insert(values=res_rev_values)
            engine.execute(insert_sql)

    package_table.c.download_url.drop()
    package_revision_table.c.download_url.drop()

def downgrade(migrate_engine):
    raise NotImplementedError()



# encoding: utf-8

from migrate import *

def upgrade(migrate_engine):
    migrate_engine.execute('''
    BEGIN;
    CREATE INDEX idx_revision_author ON "revision" (author);
    CREATE INDEX idx_openid ON "user" (openid);
    CREATE INDEX "idx_user_name_index" on "user"((CASE WHEN ("user".fullname IS NULL OR "user".fullname = '') THEN "user".name ELSE "user".fullname END));
    COMMIT;
    '''
    )


